#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Everything relating to the 'project context' object that describes the project for which the documentation is being generated.
"""

import sys
import copy
import os

try:
    import pytomlpp as toml  # fast; based on toml++ (C++)
except ImportError:
    try:
        import tomllib as toml  # PEP 680
    except ImportError:
        import tomli as toml

import datetime
import itertools
from colorama import Fore, Style

from . import doxygen, emoji, paths, repos
from .schemas import *
from .utils import *
from .version import *

# =======================================================================================================================
# internal helpers
# =======================================================================================================================


class Defaults(object):
    inline_namespaces = {r'std::(?:literals::)?(?:chrono|complex|string|string_view)_literals'}
    macros = {
        r'NDEBUG': 1,
        r'DOXYGEN': 1,
        r'__DOXYGEN__': 1,
        r'__doxygen__': 1,
        r'__POXY__': 1,
        r'__poxy__': 1,
        r'__has_include(...)': 0,
        r'__has_attribute(...)': 0,
        r'__has_builtin(...)': 0,
        r'__has_feature(...)': 0,
        r'__has_cpp_attribute(...)': 999999,
        r'POXY_IMPLEMENTATION_DETAIL(...)': r'POXY_IMPLEMENTATION_DETAIL_IMPL',
        r'POXY_IGNORE(...)': r'',
    }
    cpp_builtin_macros = {
        1998: {r'__cplusplus': r'199711L', r'__cpp_rtti': 199711, r'__cpp_exceptions': 199711},
        2003: dict(),  # apparently?
        2011: {
            r'__cplusplus': r'201103L',
            r'__cpp_unicode_characters': 200704,
            r'__cpp_raw_strings': 200710,
            r'__cpp_unicode_literals': 200710,
            r'__cpp_user_defined_literals': 200809,
            r'__cpp_threadsafe_static_init': 200806,
            r'__cpp_lambdas': 200907,
            r'__cpp_constexpr': 200704,
            r'__cpp_range_based_for': 200907,
            r'__cpp_static_assert': 200410,
            r'__cpp_decltype': 200707,
            r'__cpp_attributes': 200809,
            r'__cpp_rvalue_references': 200610,
            r'__cpp_variadic_templates': 200704,
            r'__cpp_initializer_lists': 200806,
            r'__cpp_delegating_constructors': 200604,
            r'__cpp_nsdmi': 200809,
            r'__cpp_inheriting_constructors': 200802,
            r'__cpp_ref_qualifiers': 200710,
            r'__cpp_alias_templates': 200704,
        },
        2014: {
            r'__cplusplus': r'201402L',
            r'__cpp_binary_literals': 201304,
            r'__cpp_init_captures': 201304,
            r'__cpp_generic_lambdas': 201304,
            r'__cpp_sized_deallocation': 201309,
            r'__cpp_constexpr': 201304,
            r'__cpp_decltype_auto': 201304,
            r'__cpp_return_type_deduction': 201304,
            r'__cpp_aggregate_nsdmi': 201304,
            r'__cpp_variable_templates': 201304,
            r'__cpp_lib_integer_sequence': 201304,
            r'__cpp_lib_exchange_function': 201304,
            r'__cpp_lib_tuples_by_type': 201304,
            r'__cpp_lib_tuple_element_t': 201402,
            r'__cpp_lib_make_unique': 201304,
            r'__cpp_lib_transparent_operators': 201210,
            r'__cpp_lib_integral_constant_callable': 201304,
            r'__cpp_lib_transformation_trait_aliases': 201304,
            r'__cpp_lib_result_of_sfinae': 201210,
            r'__cpp_lib_is_final': 201402,
            r'__cpp_lib_is_null_pointer': 201309,
            r'__cpp_lib_chrono_udls': 201304,
            r'__cpp_lib_string_udls': 201304,
            r'__cpp_lib_generic_associative_lookup': 201304,
            r'__cpp_lib_null_iterators': 201304,
            r'__cpp_lib_make_reverse_iterator': 201402,
            r'__cpp_lib_robust_nonmodifying_seq_ops': 201304,
            r'__cpp_lib_complex_udls': 201309,
            r'__cpp_lib_quoted_string_io': 201304,
            r'__cpp_lib_shared_timed_mutex': 201402,
        },
        2017: {
            r'__cplusplus': r'201703L',
            r'__cpp_hex_float': 201603,
            r'__cpp_inline_variables': 201606,
            r'__cpp_aligned_new': 201606,
            r'__cpp_guaranteed_copy_elision': 201606,
            r'__cpp_noexcept_function_type': 201510,
            r'__cpp_fold_expressions': 201603,
            r'__cpp_capture_star_this': 201603,
            r'__cpp_constexpr': 201603,
            r'__cpp_if_constexpr': 201606,
            r'__cpp_range_based_for': 201603,
            r'__cpp_static_assert': 201411,
            r'__cpp_deduction_guides': 201703,
            r'__cpp_nontype_template_parameter_auto': 201606,
            r'__cpp_namespace_attributes': 201411,
            r'__cpp_enumerator_attributes': 201411,
            r'__cpp_inheriting_constructors': 201511,
            r'__cpp_variadic_using': 201611,
            r'__cpp_structured_bindings': 201606,
            r'__cpp_aggregate_bases': 201603,
            r'__cpp_nontype_template_args': 201411,
            r'__cpp_template_template_args': 201611,
            r'__cpp_lib_byte': 201603,
            r'__cpp_lib_hardware_interference_size': 201703,
            r'__cpp_lib_launder': 201606,
            r'__cpp_lib_uncaught_exceptions': 201411,
            r'__cpp_lib_as_const': 201510,
            r'__cpp_lib_make_from_tuple': 201606,
            r'__cpp_lib_apply': 201603,
            r'__cpp_lib_optional': 201606,
            r'__cpp_lib_variant': 201606,
            r'__cpp_lib_any': 201606,
            r'__cpp_lib_addressof_constexpr': 201603,
            r'__cpp_lib_raw_memory_algorithms': 201606,
            r'__cpp_lib_transparent_operators': 201510,
            r'__cpp_lib_enable_shared_from_this': 201603,
            r'__cpp_lib_shared_ptr_weak_type': 201606,
            r'__cpp_lib_shared_ptr_arrays': 201611,
            r'__cpp_lib_memory_resource': 201603,
            r'__cpp_lib_boyer_moore_searcher': 201603,
            r'__cpp_lib_invoke': 201411,
            r'__cpp_lib_not_fn': 201603,
            r'__cpp_lib_void_t': 201411,
            r'__cpp_lib_bool_constant': 201505,
            r'__cpp_lib_type_trait_variable_templates': 201510,
            r'__cpp_lib_logical_traits': 201510,
            r'__cpp_lib_is_swappable': 201603,
            r'__cpp_lib_is_invocable': 201703,
            r'__cpp_lib_has_unique_object_representations': 201606,
            r'__cpp_lib_is_aggregate': 201703,
            r'__cpp_lib_chrono': 201611,
            r'__cpp_lib_execution': 201603,
            r'__cpp_lib_parallel_algorithm': 201603,
            r'__cpp_lib_to_chars': 201611,
            r'__cpp_lib_string_view': 201606,
            r'__cpp_lib_allocator_traits_is_always_equal': 201411,
            r'__cpp_lib_incomplete_container_elements': 201505,
            r'__cpp_lib_map_try_emplace': 201411,
            r'__cpp_lib_unordered_map_try_emplace': 201411,
            r'__cpp_lib_node_extract': 201606,
            r'__cpp_lib_array_constexpr': 201603,
            r'__cpp_lib_nonmember_container_access': 201411,
            r'__cpp_lib_sample': 201603,
            r'__cpp_lib_clamp': 201603,
            r'__cpp_lib_gcd_lcm': 201606,
            r'__cpp_lib_hypot': 201603,
            r'__cpp_lib_math_special_functions': 201603,
            r'__cpp_lib_filesystem': 201703,
            r'__cpp_lib_atomic_is_always_lock_free': 201603,
            r'__cpp_lib_shared_mutex': 201505,
            r'__cpp_lib_scoped_lock': 201703,
        },
        2020: {
            r'__cplusplus': r'202002L',
            r'__cpp_aggregate_paren_init': 201902,
            r'__cpp_char8_t': 201811,
            r'__cpp_concepts': 202002,
            r'__cpp_conditional_explicit': 201806,
            r'__cpp_consteval': 201811,
            r'__cpp_constexpr': 202002,
            r'__cpp_constexpr_dynamic_alloc': 201907,
            r'__cpp_constexpr_in_decltype': 201711,
            r'__cpp_constinit': 201907,
            r'__cpp_deduction_guides': 201907,
            r'__cpp_designated_initializers': 201707,
            r'__cpp_generic_lambdas': 201707,
            r'__cpp_impl_coroutine': 201902,
            r'__cpp_impl_destroying_delete': 201806,
            r'__cpp_impl_three_way_comparison': 201907,
            r'__cpp_init_captures': 201803,
            r'__cpp_lib_array_constexpr': 201811,
            r'__cpp_lib_assume_aligned': 201811,
            r'__cpp_lib_atomic_flag_test': 201907,
            r'__cpp_lib_atomic_float': 201711,
            r'__cpp_lib_atomic_lock_free_type_aliases': 201907,
            r'__cpp_lib_atomic_ref': 201806,
            r'__cpp_lib_atomic_shared_ptr': 201711,
            r'__cpp_lib_atomic_value_initialization': 201911,
            r'__cpp_lib_atomic_wait': 201907,
            r'__cpp_lib_barrier': 201907,
            r'__cpp_lib_bind_front': 201907,
            r'__cpp_lib_bit_cast': 201806,
            r'__cpp_lib_bitops': 201907,
            r'__cpp_lib_bounded_array_traits': 201902,
            r'__cpp_lib_char8_t': 201907,
            r'__cpp_lib_chrono': 201907,
            r'__cpp_lib_concepts': 202002,
            r'__cpp_lib_constexpr_algorithms': 201806,
            r'__cpp_lib_constexpr_complex': 201711,
            r'__cpp_lib_constexpr_dynamic_alloc': 201907,
            r'__cpp_lib_constexpr_functional': 201907,
            r'__cpp_lib_constexpr_iterator': 201811,
            r'__cpp_lib_constexpr_memory': 201811,
            r'__cpp_lib_constexpr_numeric': 201911,
            r'__cpp_lib_constexpr_string': 201907,
            r'__cpp_lib_constexpr_string_view': 201811,
            r'__cpp_lib_constexpr_tuple': 201811,
            r'__cpp_lib_constexpr_utility': 201811,
            r'__cpp_lib_constexpr_vector': 201907,
            r'__cpp_lib_coroutine': 201902,
            r'__cpp_lib_destroying_delete': 201806,
            r'__cpp_lib_endian': 201907,
            r'__cpp_lib_erase_if': 202002,
            r'__cpp_lib_execution': 201902,
            r'__cpp_lib_format': 201907,
            r'__cpp_lib_generic_unordered_lookup': 201811,
            r'__cpp_lib_int_pow2': 202002,
            r'__cpp_lib_integer_comparison_functions': 202002,
            r'__cpp_lib_interpolate': 201902,
            r'__cpp_lib_is_constant_evaluated': 201811,
            r'__cpp_lib_is_layout_compatible': 201907,
            r'__cpp_lib_is_nothrow_convertible': 201806,
            r'__cpp_lib_is_pointer_interconvertible': 201907,
            r'__cpp_lib_jthread': 201911,
            r'__cpp_lib_latch': 201907,
            r'__cpp_lib_list_remove_return_type': 201806,
            r'__cpp_lib_math_constants': 201907,
            r'__cpp_lib_polymorphic_allocator': 201902,
            r'__cpp_lib_ranges': 201911,
            r'__cpp_lib_remove_cvref': 201711,
            r'__cpp_lib_semaphore': 201907,
            r'__cpp_lib_shared_ptr_arrays': 201707,
            r'__cpp_lib_shift': 201806,
            r'__cpp_lib_smart_ptr_for_overwrite': 202002,
            r'__cpp_lib_source_location': 201907,
            r'__cpp_lib_span': 202002,
            r'__cpp_lib_ssize': 201902,
            r'__cpp_lib_starts_ends_with': 201711,
            r'__cpp_lib_string_view': 201803,
            r'__cpp_lib_syncbuf': 201803,
            r'__cpp_lib_three_way_comparison': 201907,
            r'__cpp_lib_to_address': 201711,
            r'__cpp_lib_to_array': 201907,
            r'__cpp_lib_type_identity': 201806,
            r'__cpp_lib_unwrap_ref': 201811,
            r'__cpp_modules': 201907,
            r'__cpp_nontype_template_args': 201911,
            r'__cpp_using_enum': 201907,
        },
        2023: {
            r'__cplusplus': r'202300L',
            r'__cpp_constexpr': 202110,
            r'__cpp_explicit_this_parameter': 202110,
            r'__cpp_if_consteval': 202106,
            r'__cpp_lib_adaptor_iterator_pair_constructor': 202106,
            r'__cpp_lib_allocate_at_least': 202106,
            r'__cpp_lib_associative_heterogeneous_erasure': 202110,
            r'__cpp_lib_byteswap': 202110,
            r'__cpp_lib_constexpr_typeinfo': 202106,
            r'__cpp_lib_format': 202110,
            r'__cpp_lib_invoke_r': 202106,
            r'__cpp_lib_is_scoped_enum': 202011,
            r'__cpp_lib_monadic_optional': 202110,
            r'__cpp_lib_move_only_function': 202110,
            r'__cpp_lib_optional': 202106,
            r'__cpp_lib_out_ptr': 202106,
            r'__cpp_lib_ranges': 202110,
            r'__cpp_lib_ranges_starts_ends_with': 202106,
            r'__cpp_lib_ranges_zip': 202110,
            r'__cpp_lib_spanstream': 202106,
            r'__cpp_lib_stacktrace': 202011,
            r'__cpp_lib_stdatomic_h': 202011,
            r'__cpp_lib_string_contains': 202011,
            r'__cpp_lib_string_resize_and_overwrite': 202110,
            r'__cpp_lib_to_underlying': 202102,
            r'__cpp_lib_variant': 202102,
            r'__cpp_lib_variant': 202106,
            r'__cpp_multidimensional_subscript': 202110,
            r'__cpp_size_t_suffix': 202011,
        },
        2026: {r'__cplusplus': r'202600L'},
        2029: {r'__cplusplus': r'202900L'},
    }
    autolinks = {
        # builtins
        r'const_cast': r'https://en.cppreference.com/w/cpp/language/const_cast',
        r'dynamic_cast': r'https://en.cppreference.com/w/cpp/language/dynamic_cast',
        r'reinterpret_cast': r'https://en.cppreference.com/w/cpp/language/reinterpret_cast',
        r'static_cast': r'https://en.cppreference.com/w/cpp/language/static_cast',
        r'(?:_Float|__fp)16s?': r'https://gcc.gnu.org/onlinedocs/gcc/Half-Precision.html',
        r'(?:_Float|__float)(128|80)s?': r'https://gcc.gnu.org/onlinedocs/gcc/Floating-Types.html',
        r'(?:wchar|char(?:8|16|32))_ts?': r'https://en.cppreference.com/w/cpp/language/types#Character_types',
        regex_trie(
            r'__cplusplus',  #
            r'__FILE__',
            r'__LINE__',
            r'__DATE__',
            r'__TIME__',
            r'__STDC__',
            r'__STDC_HOSTED__',
            r'__STDC_VERSION__',
            r'__STDC_ISO_10646__',
            r'__STDC_MB_MIGHT_NEQ_WC__',
            r'__STDCPP_THREADS__',
            r'__STDCPP_DEFAULT_NEW_ALIGNMENT__',
            r'__STDCPP_STRICT_POINTER_SAFETY__',
        ): r'https://en.cppreference.com/w/cpp/preprocessor/replace',
        # standard library
        r'std::assume_aligned(?:\(\))?': r'https://en.cppreference.com/w/cpp/memory/assume_aligned',
        r'(?:std::)?nullptr_t': r'https://en.cppreference.com/w/cpp/types/nullptr_t',
        r'(?:std::)?ptrdiff_t': r'https://en.cppreference.com/w/cpp/types/ptrdiff_t',
        r'(?:std::)?size_t': r'https://en.cppreference.com/w/cpp/types/size_t',
        r'(?:std::)?u?int(?:_fast|_least)?(?:8|16|32|64)_ts?': r'https://en.cppreference.com/w/cpp/types/integer',
        r'(?:std::)?u?int(?:max|ptr)_t': r'https://en.cppreference.com/w/cpp/types/integer',
        r'\s(?:<|&lt;)fstream(?:>|&gt;)': r'https://en.cppreference.com/w/cpp/header/fstream',
        r'\s(?:<|&lt;)iosfwd(?:>|&gt;)': r'https://en.cppreference.com/w/cpp/header/iosfwd',
        r'\s(?:<|&lt;)iostream(?:>|&gt;)': r'https://en.cppreference.com/w/cpp/header/iostream',
        r'\s(?:<|&lt;)sstream(?:>|&gt;)': r'https://en.cppreference.com/w/cpp/header/sstream',
        r'\s(?:<|&lt;)string(?:>|&gt;)': r'https://en.cppreference.com/w/cpp/header/string',
        r'\s(?:<|&lt;)string_view(?:>|&gt;)': r'https://en.cppreference.com/w/cpp/header/string_view',
        r'std::(?:basic_|w)?fstreams?': r'https://en.cppreference.com/w/cpp/io/basic_fstream',
        r'std::(?:basic_|w)?ifstreams?': r'https://en.cppreference.com/w/cpp/io/basic_ifstream',
        r'std::(?:basic_|w)?iostreams?': r'https://en.cppreference.com/w/cpp/io/basic_iostream',
        r'std::(?:basic_|w)?istreams?': r'https://en.cppreference.com/w/cpp/io/basic_istream',
        r'std::(?:basic_|w)?istringstreams?': r'https://en.cppreference.com/w/cpp/io/basic_istringstream',
        r'std::(?:basic_|w)?ofstreams?': r'https://en.cppreference.com/w/cpp/io/basic_ofstream',
        r'std::(?:basic_|w)?ostreams?': r'https://en.cppreference.com/w/cpp/io/basic_ostream',
        r'std::(?:basic_|w)?ostringstreams?': r'https://en.cppreference.com/w/cpp/io/basic_ostringstream',
        r'std::(?:basic_|w)?stringstreams?': r'https://en.cppreference.com/w/cpp/io/basic_stringstream',
        r'std::(?:basic_|w|u(?:8|16|32))?string_views?': r'https://en.cppreference.com/w/cpp/string/basic_string_view',
        r'std::(?:basic_|w|u(?:8|16|32))?strings?': r'https://en.cppreference.com/w/cpp/string/basic_string',
        r'std::[fl]?abs[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/abs',
        r'std::acos[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/acos',
        r'std::add_[lr]value_reference(?:_t)?': r'https://en.cppreference.com/w/cpp/types/add_reference',
        r'std::add_(?:cv|const|volatile)(?:_t)?': r'https://en.cppreference.com/w/cpp/types/add_cv',
        r'std::add_pointer(?:_t)?': r'https://en.cppreference.com/w/cpp/types/add_pointer',
        r'std::allocators?': r'https://en.cppreference.com/w/cpp/memory/allocator',
        r'std::arrays?': r'https://en.cppreference.com/w/cpp/container/array',
        r'std::as_(writable_)?bytes(?:\(\))?': r'https://en.cppreference.com/w/cpp/container/span/as_bytes',
        r'std::asin[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/asin',
        r'std::atan2[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/atan2',
        r'std::atan[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/atan',
        r'std::bad_alloc': r'https://en.cppreference.com/w/cpp/memory/new/bad_alloc',
        r'std::basic_ios': r'https://en.cppreference.com/w/cpp/io/basic_ios',
        r'std::bit_cast(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/bit_cast',
        r'std::bit_ceil(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/bit_ceil',
        r'std::bit_floor(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/bit_floor',
        r'std::bit_width(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/bit_width',
        r'std::bytes?': r'https://en.cppreference.com/w/cpp/types/byte',
        r'std::ceil[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/ceil',
        r'std::char_traits': r'https://en.cppreference.com/w/cpp/string/char_traits',
        r'std::chrono::durations?': r'https://en.cppreference.com/w/cpp/chrono/duration',
        r'std::clamp(?:\(\))?': r'https://en.cppreference.com/w/cpp/algorithm/clamp',
        r'std::conditional(?:_t)?': r'https://en.cppreference.com/w/cpp/types/conditional',
        r'std::cos[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/cos',
        r'std::countl_one(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/countl_one',
        r'std::countl_zero(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/countl_zero',
        r'std::countr_one(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/countr_one',
        r'std::countr_zero(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/countr_zero',
        r'std::enable_if(?:_t)?': r'https://en.cppreference.com/w/cpp/types/enable_if',
        r'std::exceptions?': r'https://en.cppreference.com/w/cpp/error/exception',
        r'std::expecteds?': r'https://en.cppreference.com/w/cpp/utility/expected',
        r'std::floor[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/floor',
        r'std::fpos': r'https://en.cppreference.com/w/cpp/io/fpos',
        r'std::has_single_bit(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/has_single_bit',
        r'std::hash': r'https://en.cppreference.com/w/cpp/utility/hash',
        r'std::initializer_lists?': r'https://en.cppreference.com/w/cpp/utility/initializer_list',
        r'std::integral_constants?': r'https://en.cppreference.com/w/cpp/types/integral_constant',
        r'std::ios(?:_base)?': r'https://en.cppreference.com/w/cpp/io/ios_base',
        r'std::is_(?:nothrow_)?convertible(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_convertible',
        r'std::is_(?:nothrow_)?invocable(?:_r)?': r'https://en.cppreference.com/w/cpp/types/is_invocable',
        r'std::is_base_of(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_base_of',
        r'std::is_constant_evaluated(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/is_constant_evaluated',
        r'std::is_enum(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_enum',
        r'std::is_floating_point(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_floating_point',
        r'std::is_integral(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_integral',
        r'std::is_pointer(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_pointer',
        r'std::is_reference(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_reference',
        r'std::is_same(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_same',
        r'std::is_signed(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_signed',
        r'std::is_unsigned(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_unsigned',
        r'std::is_void(?:_v)?': r'https://en.cppreference.com/w/cpp/types/is_void',
        r'std::launder(?:\(\))?': r'https://en.cppreference.com/w/cpp/utility/launder',
        r'std::lerp(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/lerp',
        r'std::maps?': r'https://en.cppreference.com/w/cpp/container/map',
        r'std::max(?:\(\))?': r'https://en.cppreference.com/w/cpp/algorithm/max',
        r'std::min(?:\(\))?': r'https://en.cppreference.com/w/cpp/algorithm/min',
        r'std::numeric_limits::min(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/min',
        r'std::numeric_limits::lowest(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/lowest',
        r'std::numeric_limits::max(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/max',
        r'std::numeric_limits::epsilon(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/epsilon',
        r'std::numeric_limits::round_error(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/round_error',
        r'std::numeric_limits::infinity(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/infinity',
        r'std::numeric_limits::quiet_NaN(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/quiet_NaN',
        r'std::numeric_limits::signaling_NaN(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/signaling_NaN',
        r'std::numeric_limits::denorm_min(?:\(\))?': r'https://en.cppreference.com/w/cpp/types/numeric_limits/denorm_min',
        r'std::numeric_limits': r'https://en.cppreference.com/w/cpp/types/numeric_limits',
        r'std::optionals?': r'https://en.cppreference.com/w/cpp/utility/optional',
        r'std::pairs?': r'https://en.cppreference.com/w/cpp/utility/pair',
        r'std::popcount(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/popcount',
        r'std::remove_cv(?:_t)?': r'https://en.cppreference.com/w/cpp/types/remove_cv',
        r'std::remove_reference(?:_t)?': r'https://en.cppreference.com/w/cpp/types/remove_reference',
        r'std::reverse_iterator': r'https://en.cppreference.com/w/cpp/iterator/reverse_iterator',
        r'std::runtime_errors?': r'https://en.cppreference.com/w/cpp/error/runtime_error',
        r'std::sets?': r'https://en.cppreference.com/w/cpp/container/set',
        r'std::shared_ptrs?': r'https://en.cppreference.com/w/cpp/memory/shared_ptr',
        r'std::sin[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/sin',
        r'std::spans?': r'https://en.cppreference.com/w/cpp/container/span',
        r'std::sqrt[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/sqrt',
        r'std::tan[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/tan',
        r'std::to_address(?:\(\))?': r'https://en.cppreference.com/w/cpp/memory/to_address',
        r'std::(?:true|false)_type': r'https://en.cppreference.com/w/cpp/types/integral_constant',
        r'std::trunc[fl]?(?:\(\))?': r'https://en.cppreference.com/w/cpp/numeric/math/trunc',
        r'std::tuple_element(?:_t)?': r'https://en.cppreference.com/w/cpp/utility/tuple/tuple_element',
        r'std::tuple_size(?:_v)?': r'https://en.cppreference.com/w/cpp/utility/tuple/tuple_size',
        r'std::tuples?': r'https://en.cppreference.com/w/cpp/utility/tuple',
        r'std::type_identity(?:_t)?': r'https://en.cppreference.com/w/cpp/types/type_identity',
        r'std::underlying_type(?:_t)?': r'https://en.cppreference.com/w/cpp/types/underlying_type',
        r'std::unique_ptrs?': r'https://en.cppreference.com/w/cpp/memory/unique_ptr',
        r'std::unordered_maps?': r'https://en.cppreference.com/w/cpp/container/unordered_map',
        r'std::unordered_sets?': r'https://en.cppreference.com/w/cpp/container/unordered_set',
        r'std::vectors?': r'https://en.cppreference.com/w/cpp/container/vector',
        r'std::atomic[a-zA-Z_0-9]*': r'https://en.cppreference.com/w/cpp/atomic/atomic',
        r'std::to_underlying(?:\(\))?': r'https://en.cppreference.com/w/cpp/utility/to_underlying',
        r'std::move_only_functions?': r'https://en.cppreference.com/w/cpp/utility/functional/move_only_function',
        # named requirements
        r'Allocators?': r'https://en.cppreference.com/w/cpp/named_req/Allocator',
        r'AllocatorAwareContainers?': r'https://en.cppreference.com/w/cpp/named_req/AllocatorAwareContainer',
        r'AssociativeContainers?': r'https://en.cppreference.com/w/cpp/named_req/AssociativeContainer',
        r'BasicFormatters?': r'https://en.cppreference.com/w/cpp/named_req/BasicFormatter',
        r'BasicLockables?': r'https://en.cppreference.com/w/cpp/named_req/BasicLockable',
        r'(?:Legacy)?BidirectionalIterators?': r'https://en.cppreference.com/w/cpp/named_req/BidirectionalIterator',
        r'BinaryPredicates?': r'https://en.cppreference.com/w/cpp/named_req/BinaryPredicate',
        r'BinaryTypeTraits?': r'https://en.cppreference.com/w/cpp/named_req/BinaryTypeTrait',
        r'BitmaskTypes?': r'https://en.cppreference.com/w/cpp/named_req/BitmaskType',
        r'Callables?': r'https://en.cppreference.com/w/cpp/named_req/Callable',
        r'CharTraits?': r'https://en.cppreference.com/w/cpp/named_req/CharTraits',
        r'Compares?': r'https://en.cppreference.com/w/cpp/named_req/Compare',
        r'(?:Legacy)?ConstexprIterators?': r'https://en.cppreference.com/w/cpp/named_req/ConstexprIterator',
        r'Containers?': r'https://en.cppreference.com/w/cpp/named_req/Container',
        r'ContiguousContainers?': r'https://en.cppreference.com/w/cpp/named_req/ContiguousContainer',
        r'(?:Legacy)?ContiguousIterators?': r'https://en.cppreference.com/w/cpp/named_req/ContiguousIterator',
        r'CopyAssignables?': r'https://en.cppreference.com/w/cpp/named_req/CopyAssignable',
        r'CopyConstructibles?': r'https://en.cppreference.com/w/cpp/named_req/CopyConstructible',
        r'CopyInsertables?': r'https://en.cppreference.com/w/cpp/named_req/CopyInsertable',
        r'DefaultConstructibles?': r'https://en.cppreference.com/w/cpp/named_req/DefaultConstructible',
        r'DefaultInsertables?': r'https://en.cppreference.com/w/cpp/named_req/DefaultInsertable',
        r'Destructibles?': r'https://en.cppreference.com/w/cpp/named_req/Destructible',
        r'EmplaceConstructibles?': r'https://en.cppreference.com/w/cpp/named_req/EmplaceConstructible',
        r'EqualityComparables?': r'https://en.cppreference.com/w/cpp/named_req/EqualityComparable',
        r'Erasables?': r'https://en.cppreference.com/w/cpp/named_req/Erasable',
        r'FormattedInputFunctions?': r'https://en.cppreference.com/w/cpp/named_req/FormattedInputFunction',
        r'FormattedOutputFunctions?': r'https://en.cppreference.com/w/cpp/named_req/FormattedOutputFunction',
        r'(?:Legacy)?ForwardIterators?': r'https://en.cppreference.com/w/cpp/named_req/ForwardIterator',
        r'FunctionObjects?': r'https://en.cppreference.com/w/cpp/named_req/FunctionObject',
        r'ImplicitLifetimeTypes?': r'https://en.cppreference.com/w/cpp/named_req/ImplicitLifetimeType',
        r'LegacyIterators?': r'https://en.cppreference.com/w/cpp/named_req/LegacyIterator',
        r'(?:Legacy)?InputIterators?': r'https://en.cppreference.com/w/cpp/named_req/InputIterator',
        r'(?:Legacy)?Iterators?': r'https://en.cppreference.com/w/cpp/named_req/Iterator',
        r'LessThanComparables?': r'https://en.cppreference.com/w/cpp/named_req/LessThanComparable',
        r'LiteralTypes?': r'https://en.cppreference.com/w/cpp/named_req/LiteralType',
        r'Lockables?': r'https://en.cppreference.com/w/cpp/named_req/Lockable',
        r'MoveAssignables?': r'https://en.cppreference.com/w/cpp/named_req/MoveAssignable',
        r'MoveConstructibles?': r'https://en.cppreference.com/w/cpp/named_req/MoveConstructible',
        r'MoveInsertables?': r'https://en.cppreference.com/w/cpp/named_req/MoveInsertable',
        r'NullablePointers?': r'https://en.cppreference.com/w/cpp/named_req/NullablePointer',
        r'NumericTypes?': r'https://en.cppreference.com/w/cpp/named_req/NumericType',
        r'(?:Legacy)?OutputIterators?': r'https://en.cppreference.com/w/cpp/named_req/OutputIterator',
        r'PODTypes?': r'https://en.cppreference.com/w/cpp/named_req/PODType',
        r'Predicates?': r'https://en.cppreference.com/w/cpp/named_req/Predicate',
        r'(?:Legacy)?RandomAccessIterators?': r'https://en.cppreference.com/w/cpp/named_req/RandomAccessIterator',
        r'RandomNumberDistributions?': r'https://en.cppreference.com/w/cpp/named_req/RandomNumberDistribution',
        r'RandomNumberEngines?': r'https://en.cppreference.com/w/cpp/named_req/RandomNumberEngine',
        r'RandomNumberEngineAdaptors?': r'https://en.cppreference.com/w/cpp/named_req/RandomNumberEngineAdaptor',
        r'RangeAdaptorClosureObjects?': r'https://en.cppreference.com/w/cpp/named_req/RangeAdaptorClosureObject',
        r'RangeAdaptorObjects?': r'https://en.cppreference.com/w/cpp/named_req/RangeAdaptorObject',
        r'RegexTraitss?': r'https://en.cppreference.com/w/cpp/named_req/RegexTraits',
        r'ReversibleContainers?': r'https://en.cppreference.com/w/cpp/named_req/ReversibleContainer',
        r'ScalarTypes?': r'https://en.cppreference.com/w/cpp/named_req/ScalarType',
        r'SeedSequences?': r'https://en.cppreference.com/w/cpp/named_req/SeedSequence',
        r'SequenceContainers?': r'https://en.cppreference.com/w/cpp/named_req/SequenceContainer',
        r'SharedLockables?': r'https://en.cppreference.com/w/cpp/named_req/SharedLockable',
        r'SharedMutexes?': r'https://en.cppreference.com/w/cpp/named_req/SharedMutex',
        r'SharedTimedLockables?': r'https://en.cppreference.com/w/cpp/named_req/SharedTimedLockable',
        r'SharedTimedMutexes?': r'https://en.cppreference.com/w/cpp/named_req/SharedTimedMutex',
        r'StandardLayoutTypes?': r'https://en.cppreference.com/w/cpp/named_req/StandardLayoutType',
        r'Swappables?': r'https://en.cppreference.com/w/cpp/named_req/Swappable',
        r'TimedLockables?': r'https://en.cppreference.com/w/cpp/named_req/TimedLockable',
        r'TimedMutexes?': r'https://en.cppreference.com/w/cpp/named_req/TimedMutex',
        r'TransformationTraits?': r'https://en.cppreference.com/w/cpp/named_req/TransformationTrait',
        r'TrivialClocks?': r'https://en.cppreference.com/w/cpp/named_req/TrivialClock',
        r'TriviallyCopyables?': r'https://en.cppreference.com/w/cpp/named_req/TriviallyCopyable',
        r'TrivialTypes?': r'https://en.cppreference.com/w/cpp/named_req/TrivialType',
        r'UnaryTypeTraits?': r'https://en.cppreference.com/w/cpp/named_req/UnaryTypeTrait',
        r'UnformattedInputFunctions?': r'https://en.cppreference.com/w/cpp/named_req/UnformattedInputFunction',
        r'UnformattedOutputFunctions?': r'https://en.cppreference.com/w/cpp/named_req/UnformattedOutputFunction',
        r'UniformRandomBitGenerators?': r'https://en.cppreference.com/w/cpp/named_req/UniformRandomBitGenerator',
        r'UnorderedAssociativeContainers?': r'https://en.cppreference.com/w/cpp/named_req/UnorderedAssociativeContainer',
        r'ValueSwappables?': r'https://en.cppreference.com/w/cpp/named_req/ValueSwappable',
        # other misc cppreference pages
        r'(?:RAII|[Rr]esource[ -][Aa]cquisition[ -][Ii]s[ -][Ii]nitiali[sz]ation(?: \(RAII\))?)': r'https://en.cppreference.com/w/cpp/language/raii',
        r'(?:CRTP|[Cc]uriously[ -][Rr]ecurring[ -][Tt]emplate[ -][Pp]attern(?: \(CRTP\))?)': r'https://en.wikipedia.org/wiki/Curiously_recurring_template_pattern',
        r'(?:CTAD|[Cc]lass[ -][Tt]emplate[ -][Aa]rgument[ -][Dd]eduction(?: \(CTAD\))?)': r'https://en.cppreference.com/w/cpp/language/class_template_argument_deduction',
        r'(?:EBC?O|[Ee]mpty[ -][Bb]ase[ -](?:[Cc]lass[ -])?[Oo]ptimi[sz]ations?(?: \(EBC?O\))?)': r'https://en.cppreference.com/w/cpp/language/ebo',
        r'(?:UDLs?|[Uu]ser[ -][Dd]efined[ -][Ll]iterals?(?:[ -][Oo]perators?)?(?: \(UDLs?\))?)': r'https://en.cppreference.com/w/cpp/language/user_literal',
        r'(?:N?RVO|[Cc]opy[ -][Ee]lision|(?:[Nn]amed[ -])?[Rr]eturn[ -][Vv]alue[ -][Oo]ptimi[sz]ations?(?: \(N?RVO\))?)': r'https://en.cppreference.com/w/cpp/language/copy_elision',
        r'INVOKE(?:(?:<|&lt;)R(?:>|&gt;))?(?: [Ee]xpressions?)?': r'https://en.cppreference.com/w/cpp/utility/functional',
        # windows
        r'(?:L?P)?(?:'
        + r'D?WORD(?:32|64|_PTR)?|HANDLE|HMODULE|BOOL(?:EAN)?'
        + r'|U?SHORT|U?LONG|U?INT(?:8|16|32|64)?'
        + r'|BYTE|VOID|C[WT]?STR'
        + r')': r'https://docs.microsoft.com/en-us/windows/desktop/winprog/windows-data-types',
        r'__INTELLISENSE__|_MSC(?:_FULL)_VER|_MSVC_LANG|_WIN(?:32|64)': r'https://docs.microsoft.com/en-us/cpp/preprocessor/predefined-macros?view=vs-2019',
        r'IUnknowns?': r'https://docs.microsoft.com/en-us/windows/win32/api/unknwn/nn-unknwn-iunknown',
        r'(?:IUnknown::)?QueryInterface?': r'https://docs.microsoft.com/en-us/windows/win32/api/unknwn/nf-unknwn-iunknown-queryinterface(q)',
        # unreal engine types
        r'(?:::)?FBox(?:es)?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FBox/index.html',
        r'(?:::)?FBox2Ds?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FBox2D/index.html',
        r'(?:::)?FColors?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FColor/index.html',
        r'(?:::)?FFloat16s?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FFloat16/index.html',
        r'(?:::)?FFloat32s?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FFloat32/index.html',
        r'(?:::)?FMatrix(?:es|ices)?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FMatrix/index.html',
        r'(?:::)?FMatrix2x2s?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FMatrix2x2/index.html',
        r'(?:::)?FOrientedBox(?:es)?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FOrientedBox/index.html',
        r'(?:::)?FPlanes?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FPlane/index.html',
        r'(?:::)?FQuat2Ds?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FQuat2D/index.html',
        r'(?:::)?FQuats?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FQuat/index.html',
        r'(?:::)?FStrings?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Containers/FString/index.html',
        r'(?:::)?FVector2DHalfs?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FVector2DHalf/index.html',
        r'(?:::)?FVector2Ds?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FVector2D/index.html',
        r'(?:::)?FVector4s?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FVector4/index.html',
        r'(?:::)?FVectors?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/FVector/index.html',
        r'(?:::)?TMatrix(?:es|ices)?': r'https://docs.unrealengine.com/en-US/API/Runtime/Core/Math/TMatrix/index.html',
        r'(?:::)?UStaticMesh(?:es)?': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/Engine/UStaticMesh/',
        r'(?:::)?FStaticMeshLODResources': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/FStaticMeshLODResources/',
        r'(?:::)?FRawStaticIndexBuffer(?:s)?': r'https://docs.unrealengine.com/4.26/en-US/API/Runtime/Engine/FRawStaticIndexBuffer/',
        r'(?:::)?FStaticMeshVertexBuffers': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/FStaticMeshVertexBuffers/',
        r'(?:::)?FStaticMeshVertexBuffer': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/Rendering/FStaticMeshVertexBuffer/',
        r'(?:::)?FPositionVertexBuffer(?:s)?': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Engine/Rendering/FPositionVertexBuffer/',
        r'(?:::)?TArrayView(?:s)?': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Core/Containers/TArrayView/',
        r'(?:::)?TArray(?:s)?': r'https://docs.unrealengine.com/4.27/en-US/API/Runtime/Core/Containers/TArray/',
        # fmtlib
        r'fmt::(?:v?format(?:_string|_to|ted_size|ter)?|memory_buffer|string_view|runtime|print|to_w?string|underlying|ptr|emphasis|styled)s?': 'https://fmt.dev/latest/api.html',
    }
    navbar = (r'files', r'groups', r'namespaces', r'classes', r'concepts')
    navbar_all = (r'pages', *navbar, r'repo', r'theme')
    aliases = {
        # poxy
        r'cpp': r'@code{.cpp}',
        r'ecpp': r'@endcode',
        r'endcpp': r'@endcode',
        r'out': r'@code{.shell-session}',
        r'eout': r'@endcode',
        r'endout': r'@endcode',
        r'python': r'@code{.py}',
        r'epython': r'@endcode',
        r'endpython': r'@endcode',
        r'meson': r'@code{.py}',
        r'emeson': r'@endcode',
        r'endmeson': r'@endcode',
        r'cmake': r'@code{.cmake}',
        r'ecmake': r'@endcode',
        r'endcmake': r'@endcode',
        r'javascript': r'@code{.js}',
        r'ejavascript': r'@endcode',
        r'endjavascript': r'@endcode',
        r'json': r'@code{.js}',
        r'ejson': r'@endcode',
        r'endjson': r'@endcode',
        r'shell': r'@code{.shell-session}',
        r'eshell': r'@endcode',
        r'endshell': r'@endcode',
        r'bash': r'@code{.sh}',
        r'ebash': r'@endcode',
        r'endbash': r'@endcode',
        r'detail': r'@details',
        r'conditional_return{1}': r'<strong><em>\1:</em></strong> ',
        r'inline_attention': r'[set_class m-note m-warning]',
        r'inline_note': r'[set_class m-note m-info]',
        r'inline_remark': r'[set_class m-note m-default]',
        r'inline_subheading{1}': r'[h4]\1[/h4]',
        r'inline_success': r'[set_class m-note m-success]',
        r'inline_warning': r'[set_class m-note m-danger]',
        r'github{1}': r'<a href="https://github.com/\1" target="_blank">\1</a>',
        r'github{2}': r'<a href="https://github.com/\1" target="_blank">\2</a>',
        r'gitlab{1}': r'<a href="https://gitlab.com/\1" target="_blank">\1</a>',
        r'gitlab{2}': r'<a href="https://gitlab.com/\1" target="_blank">\2</a>',
        r'godbolt{1}': r'<a href="https://godbolt.org/z/\1" target="_blank">Try this code on Compiler Explorer</a>',
        r'flags_enum': r'@note This enum is a flags type; it is equipped with a full complement of bitwise operators. ^^',
        r'implementers': r'@par [parent_parent_set_class m-block m-dim][emoji hammer][entity nbsp]Implementers: ',
        r'optional': r'@par [parent_parent_set_class m-block m-info]Optional field ^^',
        r'required': r'@par [parent_parent_set_class m-block m-warning][emoji warning][entity nbsp]Required field ^^',
        r'availability': r'@par [parent_parent_set_class m-block m-special]Conditional availability ^^',
        r'figure{1}': r'@image html \1',
        r'figure{2}': r'@image html \1 "\2"',
        # m.css
        r'm_div{1}': r'@xmlonly<mcss:div xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:class="\1">@endxmlonly',
        r'm_enddiv': r'@xmlonly</mcss:div>@endxmlonly',
        r'm_span{1}': r'@xmlonly<mcss:span xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:class="\1">@endxmlonly',
        r'm_endspan': r'@xmlonly</mcss:span>@endxmlonly',
        r'm_class{1}': r'@xmlonly<mcss:class xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:class="\1" />@endxmlonly',
        r'm_footernavigation': r'@xmlonly<mcss:footernavigation xmlns:mcss="http://mcss.mosra.cz/doxygen/" />@endxmlonly',
        r'm_examplenavigation{2}': r'@xmlonly<mcss:examplenavigation xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:page="\1" mcss:prefix="\2" />@endxmlonly',
        r'm_keywords{1}': r'@xmlonly<mcss:search xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:keywords="\1" />@endxmlonly',
        r'm_keyword{3}': r'@xmlonly<mcss:search xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:keyword="\1" mcss:title="\2" mcss:suffix-length="\3" />@endxmlonly',
        r'm_enum_values_as_keywords': r'@xmlonly<mcss:search xmlns:mcss="http://mcss.mosra.cz/doxygen/" mcss:enum-values-as-keywords="true" />@endxmlonly',
    }
    source_patterns = {
        r'*.h',
        r'*.hh',
        r'*.hxx',
        r'*.hpp',
        r'*.h++',
        r'*.ixx',
        r'*.inc',
        r'*.markdown',
        r'*.md',
        r'*.dox',
    }
    # code block syntax highlighting only
    #
    # note: we don't need to be comprehensive for the std namespace symbols;
    # those will get added from the cppreference tagfile
    cb_enums = {r'(?:std::)?ios(?:_base)?::(?:app|binary|in|out|trunc|ate)'}
    cb_macros = {
        # standard builtins:
        r'__cplusplus(?:_cli|_winrt)?',
        r'__cpp_[a-zA-Z_]+?',
        r'__has_(?:(?:cpp_)?attribute|include)',
        r'assert',
        r'offsetof',
        # poxy:
        r'POXY_[a-zA-Z_]+',
        # compiler builtins:
        regex_trie(
            # standard:
            r'__FILE__',
            r'__LINE__',
            r'__DATE__',
            r'__TIME__',
            r'__STDC__',
            r'__STDC_HOSTED__',
            r'__STDC_VERSION__',
            r'__STDC_ISO_10646__',
            r'__STDC_MB_MIGHT_NEQ_WC__',
            r'__STDCPP_THREADS__',
            r'__STDCPP_DEFAULT_NEW_ALIGNMENT__',
            r'__STDCPP_STRICT_POINTER_SAFETY__',
            # msvc:
            r'_ATL_VER',  #
            r'_CHAR_UNSIGNED',
            r'_CONTROL_FLOW_GUARD',
            r'_CPPRTTI',
            r'_CPPUNWIND',
            r'_DEBUG',
            r'_DLL',
            r'_INTEGRAL_MAX_BITS',
            r'_ISO_VOLATILE',
            r'_KERNEL_MODE',
            r'_MANAGED',
            r'_MFC_VER',
            r'_MSC_BUILD',
            r'_MSC_EXTENSIONS',
            r'_MSC_FULL_VER',
            r'_MSC_VER',
            r'_MSVC_LANG',
            r'_MSVC_TRADITIONAL',
            r'_MT',
            r'_M_AMD64',
            r'_M_ARM',
            r'_M_ARM64',
            r'_M_ARM64EC',
            r'_M_ARM_ARMV7VE',
            r'_M_ARM_FP',
            r'_M_CEE',
            r'_M_CEE_PURE',
            r'_M_CEE_SAFE',
            r'_M_FP_CONTRACT',
            r'_M_FP_EXCEPT',
            r'_M_FP_FAST',
            r'_M_FP_PRECISE',
            r'_M_FP_STRICT',
            r'_M_IX86',
            r'_M_IX86_FP',
            r'_M_X64',
            r'_NATIVE_WCHAR_T_DEFINED',
            r'_OPENMP',
            r'_PREFAST_',
            r'_VC_NODEFAULTLIB',
            r'_WCHAR_T_DEFINED',
            r'_WIN32',
            r'_WIN64',
            r'_WINRT_DLL',
            r'__ATOM__',
            r'__AVX2__',
            r'__AVX512BW__',
            r'__AVX512CD__',
            r'__AVX512DQ__',
            r'__AVX512F__',
            r'__AVX512VL__',
            r'__AVX__',
            r'__CLR_VER',
            r'__COUNTER__',
            r'__FUNCDNAME__',
            r'__FUNCSIG__',
            r'__FUNCTION__',
            r'__INTELLISENSE__',
            r'__MSVC_RUNTIME_CHECKS',
            r'__SANITIZE_ADDRESS__',
            r'__STDC_NO_ATOMICS__',
            r'__STDC_NO_COMPLEX__',
            r'__STDC_NO_THREADS__',
            r'__STDC_NO_VLA__',
            r'__TIMESTAMP__',
            r'__cplusplus_cli',
            r'__cplusplus_winrt',
            r'NDEBUG',
            r'_DEBUG',
        ),
    }
    cb_namespaces = {
        regex_trie(
            r'std',
            r'std::chrono',
            r'std::execution',
            r'std::filesystem',
            r'std::experimental',
            r'std::numbers',
            r'std::literals',
            r'std::literals::chrono_literals',
            r'std::literals::complex_literals',
            r'std::literals::string_literals',
            r'std::literals::string_view_literals',
            r'std::chrono_literals',
            r'std::complex_literals',
            r'std::string_literals',
            r'std::string_view_literals',
            r'std::ranges',
            r'std::this_thread',
        )
    }
    cb_types = {
        # ------ built-in types
        r'__(?:float|fp)[0-9]{1,3}',
        r'__m[0-9]{1,3}[di]?',
        r'_Float[0-9]{1,3}',
        r'[a-zA-Z_][a-zA-Z_0-9]*_t(?:ype(?:def)?|raits)?',
        regex_trie(r'bool', r'char', r'double', r'float', r'int', r'long', r'short', r'signed', r'unsigned'),
        # ------ documentation-only types
        r'(?:[a-zA-Z][a-zA-Z_]+::)*?(?:[a-zA-Z][a-zA-Z_]+_type|Foo|Bar|[Vv]ec(?:tor)?[1-4][hifd]?|[Mm]at(?:rix)?[1-4](?:[xX][1-4])?[hifd]?)',
        r'[S-Z][0-9]?',
    }
    cb_functions = {regex_trie(r'std::as_const', r'std::move', r'std::forward')}


def extract_kvps(
    config,
    table,
    key_getter=str,
    value_getter=str,
    strip_keys=True,
    strip_values=True,
    allow_blank_keys=False,
    allow_blank_values=False,
    value_type=None,
):
    assert config is not None
    assert isinstance(config, dict)
    assert table is not None

    if table not in config:
        return {}

    out = {}
    for k, v in config[table].items():
        key = key_getter(k)
        if isinstance(key, str):
            if strip_keys:
                key = key.strip()
            if not allow_blank_keys and not key:
                raise Error(rf'{table}: keys cannot be blank')
        if key in out:
            raise Error(rf'{table}.{key}: cannot be specified more than once')

        value = value_getter(v)
        if isinstance(value, str):
            if strip_values and isinstance(value, str):
                value = value.strip()
            if not allow_blank_values and not value:
                raise Error(rf'{table}.{key}: values cannot be blank')

        if value_type is not None:
            value = value_type(value)

        out[key] = value

    return out


def assert_no_unexpected_keys(raw, validated, prefix=''):
    for key in raw:
        if key not in validated:
            raise Error(rf"Unknown config property '{prefix}{key}'")
        if isinstance(validated[key], dict):
            assert_no_unexpected_keys(raw[key], validated[key], prefix=rf'{prefix}{key}.')
    return validated


class Warnings(object):
    schema = {Optional(r'enabled'): bool, Optional(r'treat_as_errors'): bool, Optional(r'undocumented'): bool}

    def __init__(self, config):
        self.enabled = True
        self.treat_as_errors = False
        self.undocumented = False

        if config is None or 'warnings' not in config:
            return

        config = config['warnings']
        if r'enabled' in config:
            self.enabled = bool(config[r'enabled'])
        if r'treat_as_errors' in config:
            self.treat_as_errors = bool(config[r'treat_as_errors'])
        if r'undocumented' in config:
            self.undocumented = bool(config[r'undocumented'])


class CodeBlocks(object):
    schema = {
        Optional(r'types'): ValueOrArray(str, name=r'types'),
        Optional(r'macros'): ValueOrArray(str, name=r'macros'),
        Optional(r'string_literals'): ValueOrArray(str, name=r'string_literals'),  # deprecated
        Optional(r'numeric_literals'): ValueOrArray(str, name=r'numeric_literals'),  # deprecated
        Optional(r'enums'): ValueOrArray(str, name=r'enums'),
        Optional(r'namespaces'): ValueOrArray(str, name=r'namespaces'),
        Optional(r'functions'): ValueOrArray(str, name=r'functions'),
    }

    def __init__(self, config, macros):
        self.types = copy.deepcopy(Defaults.cb_types)
        self.macros = copy.deepcopy(Defaults.cb_macros)
        self.enums = copy.deepcopy(Defaults.cb_enums)
        self.namespaces = copy.deepcopy(Defaults.cb_namespaces)
        self.functions = copy.deepcopy(Defaults.cb_functions)

        if r'code_blocks' in config:
            config = config['code_blocks']

            if 'types' in config:
                for t in coerce_collection(config['types']):
                    type_ = t.strip()
                    if type_:
                        self.types.add(type_)

            if 'macros' in config:
                for m in coerce_collection(config['macros']):
                    macro = m.strip()
                    if macro:
                        self.macros.add(macro)

            if 'enums' in config:
                for e in coerce_collection(config['enums']):
                    enum = e.strip()
                    if enum:
                        self.enums.add(enum)

            if 'namespaces' in config:
                for ns in coerce_collection(config['namespaces']):
                    namespace = ns.strip()
                    if namespace:
                        self.namespaces.add(namespace)

            if 'functions' in config:
                for f in coerce_collection(config['functions']):
                    function = f.strip()
                    if function:
                        self.functions.add(function)

        for k, v in macros.items():
            define = k
            bracket = define.find('(')
            if bracket != -1:
                define = define[:bracket].strip()
            if define:
                self.macros.add(define)


class Inputs(object):
    schema = {
        Optional(r'paths'): ValueOrArray(str, name=r'paths'),
        Optional(r'recursive_paths'): ValueOrArray(str, name=r'recursive_paths'),
        Optional(r'ignore'): ValueOrArray(str, name=r'ignore'),
    }

    def __init__(self, config, key, input_dir, additional_inputs=None, additional_recursive_inputs=None):
        self.paths = []

        if key in config:
            config = config[key]
        else:
            config = None

        all_paths = set()
        for recursive in (False, True):
            key = r'recursive_paths' if recursive else r'paths'
            paths = []
            if not recursive and additional_inputs is not None:
                paths = paths + [p for p in coerce_collection(additional_inputs) if p is not None]
            if recursive and additional_recursive_inputs is not None:
                paths = paths + [p for p in coerce_collection(additional_recursive_inputs) if p is not None]
            if config is not None and key in config:
                paths = paths + [p for p in coerce_collection(config[key]) if p is not None]
            paths = [p for p in paths if p]
            paths = [str(p).strip().replace('\\', '/') for p in paths]
            paths = [Path(p) for p in paths if p]
            paths = [Path(input_dir, p) if not p.is_absolute() else p for p in paths]
            paths = [p.resolve() for p in paths]
            for path in paths:
                if not path.exists():
                    raise Error(rf"{key}: '{path}' does not exist")
                if not (path.is_file() or path.is_dir()):
                    raise Error(rf"{key}: '{path}' was not a directory or file")
                all_paths.add(path)
                if recursive and path.is_dir():
                    for subdir in enumerate_directories(
                        path, filter=lambda p: not p.name.startswith(r'.'), recursive=True
                    ):
                        all_paths.add(subdir)

        ignores = set()
        if config is not None and r'ignore' in config:
            for s in coerce_collection(config[r'ignore']):
                ignore = s.strip()
            ignores = [re.compile(i) for i in ignores]
        for ignore in ignores:
            all_paths = [p for p in all_paths if not ignore.search(str(p))]

        self.paths = list(all_paths)
        self.paths.sort()


class FilteredInputs(Inputs):
    schema = combine_dicts(Inputs.schema, {Optional(r'patterns'): ValueOrArray(str, name=r'patterns')})

    def __init__(self, config, key, input_dir, additional_inputs=None, additional_recursive_inputs=None):
        super().__init__(
            config,
            key,
            input_dir,
            additional_inputs=additional_inputs,
            additional_recursive_inputs=additional_recursive_inputs,
        )
        self.patterns = None

        if key not in config:
            return
        config = config[key]

        if r'patterns' in config:
            self.patterns = set()
            for v in coerce_collection(config[r'patterns']):
                val = v.strip()
                if val:
                    self.patterns.add(val)


class Sources(FilteredInputs):
    schema = combine_dicts(
        FilteredInputs.schema,
        {
            Optional(r'strip_paths'): ValueOrArray(str, name=r'strip_paths'),
            Optional(r'strip_includes'): ValueOrArray(str, name=r'strip_includes'),
            Optional(r'extract_all'): bool,
        },
    )

    def __init__(
        self,
        config,
        key,
        input_dir,
        additional_inputs=None,
        additional_recursive_inputs=None,
        additional_strip_paths=None,
    ):
        super().__init__(
            config,
            key,
            input_dir,
            additional_inputs=additional_inputs,
            additional_recursive_inputs=additional_recursive_inputs,
        )

        self.strip_paths = []
        self.strip_includes = []
        self.extract_all = False
        if self.patterns is None:
            self.patterns = copy.deepcopy(Defaults.source_patterns)

        if key not in config:
            return
        config = config[key]

        strip_path_sources = (
            coerce_collection(config[r'strip_paths']) if r'strip_paths' in config else None,
            coerce_collection(additional_strip_paths) if additional_strip_paths is not None else None,
        )
        for sps in strip_path_sources:
            if sps is None:
                continue
            for s in sps:
                if s is None:
                    continue
                path = str(s).strip()
                if path:
                    self.strip_paths.append(path)

        if r'strip_includes' in config:
            for s in coerce_collection(config[r'strip_includes']):
                path = s.strip().replace('\\', '/')
                if path:
                    self.strip_includes.append(path)
            self.strip_includes.sort(key=lambda v: len(v), reverse=True)

        if r'extract_all' in config:
            self.extract_all = bool(config['extract_all'])


# =======================================================================================================================
# project context
# =======================================================================================================================


class Context(object):
    """
    The context object passed around during one invocation.
    """

    __config_schema = Schema(
        {
            Optional(r'aliases'): {str: str},
            Optional(r'author'): Stripped(str),
            Optional(r'autolinks'): {str: str},
            Optional(r'badges'): {str: ValueOrArray(str, name=r'badges', length=2)},
            Optional(r'changelog'): Or(str, bool),
            Optional(r'main_page'): Or(str, bool),
            Optional(r'code_blocks'): CodeBlocks.schema,
            Optional(r'cpp'): Or(str, int, error=r'cpp: expected string or integer'),
            Optional(r'defines'): {str: Or(str, int, bool)},  # legacy
            Optional(r'description'): Stripped(str),
            Optional(r'examples'): FilteredInputs.schema,
            Optional(r'extra_files'): ValueOrArray(str, name=r'extra_files'),
            Optional(r'excluded_symbols'): ValueOrArray(str, name=r'excluded_symbols'),
            Optional(r'favicon'): Stripped(str),
            Optional(r'generate_tagfile'): bool,
            Optional(r'github'): Stripped(str),
            Optional(r'gitlab'): Stripped(str),
            Optional(r'twitter'): Stripped(str),
            Optional(r'sponsor'): Stripped(str),
            Optional(r'html_header'): Stripped(str),
            Optional(r'images'): Inputs.schema,
            Optional(r'implementation_headers'): {str: ValueOrArray(str)},
            Optional(r'inline_namespaces'): ValueOrArray(str, name=r'inline_namespaces'),
            Optional(r'internal_docs'): bool,
            Optional(r'jquery'): bool,
            Optional(r'license'): ValueOrArray(str, length=2, name=r'license'),
            Optional(r'logo'): Stripped(str),
            Optional(r'macros'): {str: Or(str, int, bool)},
            Optional(r'meta_tags'): {str: Or(str, int)},
            Optional(r'name'): Stripped(str),
            Optional(r'navbar'): ValueOrArray(str, name=r'navbar'),
            Optional(r'private_repo'): bool,
            Optional(r'robots'): bool,
            Optional(r'scripts'): ValueOrArray(str, name=r'scripts'),
            Optional(r'show_includes'): bool,
            Optional(r'sources'): Sources.schema,
            Optional(r'stylesheets'): ValueOrArray(str, name=r'stylesheets'),
            Optional(r'tagfiles'): {str: str},
            Optional(r'theme'): Or(r'dark', r'light', r'custom'),
            Optional(r'warnings'): Warnings.schema,
        },
        ignore_extra_keys=True,
    )
    __namespace_qualified = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*::.+$")

    def is_verbose(self):
        return self.__verbose

    def __log(self, level, msg, indent=None):
        if msg is None:
            return
        msg = str(msg).strip(r'\r\n\v\f')
        if not msg:
            return
        if indent is not None:
            indent = str(indent)
        if indent:
            with io.StringIO() as buf:
                for line in msg.splitlines():
                    print(rf'{indent}{line}', file=buf, end='\n')
                log(self.logger, buf.getvalue(), level=level)
        else:
            log(self.logger, msg, level=level)

    def verbose(self, msg, indent=None):
        if self.__verbose:
            self.__log(logging.DEBUG, msg, indent=indent)

    def info(self, msg, indent=None):
        self.__log(logging.INFO, msg, indent=indent)

    def warning(self, msg, indent=None):
        if self.warnings.treat_as_errors:
            raise WarningTreatedAsError(msg)
        else:
            self.__log(logging.WARNING, rf'{Style.BRIGHT}{Fore.YELLOW}warning:{Style.RESET_ALL} {msg}', indent=indent)

    def verbose_value(self, name, val):
        if not self.__verbose:
            return
        with io.StringIO() as buf:
            print(rf'{name+": ":<35}', file=buf, end='')
            if val is not None:
                if isinstance(val, dict):
                    if val:
                        rpad = 0
                        for k in val:
                            rpad = max(rpad, len(str(k)))
                        first = True
                        for k, v in val.items():
                            if not first:
                                print(f'\n{" ":<35}', file=buf, end='')
                            first = False
                            print(rf'{str(k):<{rpad}} => {v}', file=buf, end='')
                elif is_collection(val):
                    if val:
                        first = True
                        for v in val:
                            if not first:
                                print(f'\n{" ":<35}', file=buf, end='')
                            first = False
                            print(v, file=buf, end='')
                else:
                    print(val, file=buf, end='')
            self.verbose(buf.getvalue())

    def verbose_object(self, name, obj):
        if not self.__verbose:
            return
        if isinstance(obj, (tuple, list, dict)):
            self.verbose_value(name, obj)
        else:
            for k, v in obj.__dict__.items():
                self.verbose_value(rf'{name}.{k}', v)

    def __init__(
        self,  #
        config_path: Path,
        output_dir: Path,
        output_html: bool,
        output_xml: bool,
        threads: int,
        cleanup: bool,
        verbose: bool,
        logger,
        html_include: str,
        html_exclude: str,
        treat_warnings_as_errors: bool,
        theme: str,
        copy_assets: bool,
        temp_dir: Path = None,
        copy_config_to: Path = None,
        versions_in_navbar: bool = False,
        **kwargs,
    ):
        self.logger = logger
        self.__verbose = bool(verbose)
        self.output_html = bool(output_html)
        self.output_xml = bool(output_xml)
        self.cleanup = bool(cleanup)
        self.copy_assets = bool(copy_assets)
        self.verbose_logger = logger if self.__verbose else None
        self.versions_in_navbar = bool(versions_in_navbar)

        self.verbose_value(r'Context.output_html', self.output_html)
        self.verbose_value(r'Context.output_xml', self.output_xml)
        self.verbose_value(r'Context.cleanup', self.cleanup)

        threads = int(threads) if threads is not None else 0
        if threads <= 0:
            threads = os.cpu_count()
        self.threads = max(1, min(os.cpu_count(), threads))
        self.verbose_value(r'Context.threads', self.threads)

        # additional kwargs (experimental stuff etc)
        self.xml_v2 = bool(kwargs[r'xml_v2']) if r'xml_v2' in kwargs else False
        self.verbose_value(r'Context.xml_v2', self.xml_v2)

        # these are overridden/initialized elsewhere; they're here so duck-typing still quacks
        self.fixers = []
        self.compound_pages = dict()
        self.compound_kinds = set()

        # initial warning state
        # note that this is overwritten after the config is read;
        # it is set here first so that we can have correct 'treat_as_errors' behaviour if we add any pre-config warnings
        self.warnings = Warnings(None)
        if treat_warnings_as_errors is not None:
            self.warnings.treat_as_errors = bool(treat_warnings_as_errors)

        if html_include is not None:
            html_include = re.compile(str(html_include))
        self.html_include = html_include
        if html_exclude is not None:
            html_exclude = re.compile(str(html_exclude))
        self.html_exclude = html_exclude

        if sys.version_info >= (3, 11):
            self.now = datetime.datetime.now(datetime.UTC).replace(microsecond=0)
        else:
            self.now = datetime.datetime.utcnow().replace(microsecond=0, tzinfo=datetime.timezone.utc)

        self.verbose_value(r'dirs.PACKAGE', paths.PACKAGE)
        self.verbose_value(r'dirs.CSS', paths.CSS)
        self.verbose_value(r'dirs.GENERATED', paths.GENERATED)
        self.verbose_value(r'dirs.FONTS', paths.FONTS)
        self.verbose_value(r'dirs.IMG', paths.IMG)
        self.verbose_value(r'dirs.JS', paths.JS)
        self.verbose_value(r'dirs.MCSS', paths.MCSS)
        self.verbose_value(r'dirs.TEMP', paths.TEMP)
        self.verbose_value(r'doxygen.path()', doxygen.path())

        # resolve paths
        if 1:
            # output
            if output_dir is None:
                output_dir = Path.cwd()
            self.output_dir = coerce_path(output_dir).resolve()
            self.verbose_value(r'Context.output_dir', self.output_dir)
            assert self.output_dir.is_absolute()

            # config path
            self.config_path = Path(r'poxy.toml').resolve()
            if config_path is not None:
                self.config_path = coerce_path(config_path).resolve()
                if self.config_path.is_dir():
                    for candidate in (r'poxy.toml', r'docs/poxy.toml', r'doc/poxy.toml', r'doxygen/poxy.toml'):
                        candidate_path = self.config_path / candidate
                        if candidate_path and candidate_path.is_file():
                            self.config_path = candidate_path
                            break
                if not self.config_path.is_file():
                    raise Error(rf"Config '{self.config_path}' did not exist or was not a file")
            if copy_config_to is not None and self.config_path.is_file():
                copy_file(self.config_path, copy_config_to)
            assert self.config_path.is_absolute()
            self.verbose_value(r'Context.config_path', self.config_path)

            # input dir
            self.input_dir = self.config_path.parent
            self.verbose_value(r'Context.input_dir', self.input_dir)
            assert_existing_directory(self.input_dir)
            assert self.input_dir.is_absolute()

            # root temp dir for this run
            if temp_dir is not None:
                self.temp_dir = Path(temp_dir).absolute()
            else:
                self.temp_dir = paths.TEMP / 'contexts' / temp_dir_name_for(self.input_dir)
            self.verbose_value(r'Context.temp_dir', self.temp_dir)
            assert self.temp_dir.is_absolute()

            # temp pages dir
            self.temp_pages_dir = Path(self.temp_dir, r'pages')
            self.verbose_value(r'Context.pages_dir', self.temp_pages_dir)
            assert self.temp_pages_dir.is_absolute()

            # temp xml output path used by doxygen
            self.temp_xml_dir = Path(self.temp_dir, r'xml')
            self.temp_original_xml_dir = Path(self.temp_dir, r'xml_original')
            self.verbose_value(r'Context.temp_xml_dir', self.temp_xml_dir)
            self.verbose_value(r'Context.temp_original_xml_dir', self.temp_original_xml_dir)
            assert self.temp_xml_dir.is_absolute()
            assert self.temp_original_xml_dir.is_absolute()

            # xml output path (--xml)
            self.xml_dir = Path(self.output_dir, r'xml')
            self.verbose_value(r'Context.xml_dir', self.xml_dir)
            assert self.xml_dir.is_absolute()

            # html output path (--html)
            self.html_dir = Path(self.output_dir, r'html')
            self.verbose_value(r'Context.html_dir', self.html_dir)
            assert self.html_dir.is_absolute()

            # assets subdir in html output
            self.assets_dir = Path(self.html_dir, r'poxy')
            self.verbose_value(r'Context.assets_dir', self.assets_dir)
            assert self.assets_dir.is_absolute()

            # blog dir
            self.blog_dir = Path(self.input_dir, r'blog')
            self.verbose_value(r'Context.blog_dir', self.blog_dir)
            assert self.blog_dir.is_absolute()

            # delete leftovers from previous run and initialize temp dirs
            delete_directory(self.temp_dir, logger=self.verbose_logger)
            delete_directory(self.xml_dir, logger=self.verbose_logger)
            delete_directory(self.html_dir, logger=self.verbose_logger)
            self.temp_dir.mkdir(exist_ok=True, parents=True)
            self.temp_pages_dir.mkdir(exist_ok=True, parents=True)

            # temp doxyfile path
            self.doxyfile_path = Path(self.temp_dir, rf'Doxyfile')
            self.verbose_value(r'Context.doxyfile_path', self.doxyfile_path)
            assert self.doxyfile_path.is_absolute()

            # temp m.css config path
            self.mcss_conf_path = Path(self.temp_dir, r'conf.py')
            self.verbose_value(r'Context.mcss_conf_path', self.mcss_conf_path)
            assert self.mcss_conf_path.is_absolute()

            # misc
            self.cppref_tagfile = coerce_path(paths.PACKAGE, r'cppreference-doxygen-web.tag.xml').resolve()
            self.verbose_value(r'Context.cppref_tagfile', self.cppref_tagfile)
            assert_existing_file(self.cppref_tagfile)
            assert self.cppref_tagfile.is_absolute()

        # read + check config
        if 1:
            extra_files = []
            badges = []
            self.scripts = []
            self.stylesheets = []

            def add_internal_asset(p) -> str:
                nonlocal extra_files
                nonlocal self
                assert p is not None
                p = coerce_path(p)
                if self.copy_assets:
                    if not p.is_absolute():
                        for dir in (paths.FONTS, paths.GENERATED, paths.JS, paths.IMG):
                            new_p = dir / p
                            if new_p.exists():
                                p = new_p
                                break
                    assert_existing_file(p)
                    extra_files.append((p, rf'poxy/{p.name}'))
                return rf'poxy/{p.name}'

            config = dict()
            if self.config_path.exists():
                assert_existing_file(self.config_path)
                config = toml.loads(read_all_text_from_file(self.config_path, logger=self.logger))
            config = assert_no_unexpected_keys(config, self.__config_schema.validate(config))

            self.warnings = Warnings(config)
            if treat_warnings_as_errors is not None:
                self.warnings.treat_as_errors = bool(treat_warnings_as_errors)
            self.verbose_value(r'Context.warnings', self.warnings)

            # project name (PROJECT_NAME)
            self.name = ''
            if 'name' in config:
                self.name = config['name'].strip()
            self.verbose_value(r'Context.name', self.name)

            # project author
            self.author = ''
            if 'author' in config:
                self.author = config['author'].strip()
            self.verbose_value(r'Context.author', self.author)

            # project description (PROJECT_BRIEF)
            self.description = ''
            if 'description' in config:
                self.description = config['description'].strip()
            self.verbose_value(r'Context.description', self.description)

            # project license
            self.license = None
            if 'license' in config:
                config['license'] = coerce_collection(config['license'])
                spdx = config['license'][0].strip(" \t-._:")
                uri = config['license'][1].strip() if len(config['license']) == 2 else ''
                if spdx:
                    self.license = {r'spdx': spdx, r'uri': uri}
                if self.license:
                    badge = re.sub(r'(?:[.]0+)+$', '', spdx.lower())  # trailing .0, .0.0 etc
                    badge = badge.strip(' \t-._:')  # leading + trailing junk
                    badge = re.sub(r'[:;!@#$%^&*\\|/,.<>?`~\[\]{}()_+\-= \t]+', '_', badge)  # internal junk
                    badge = Path(paths.IMG, rf'poxy-badge-license-{badge}.svg')
                    self.verbose(rf"Finding badge SVG for license '{spdx}'...")
                    if badge.exists():
                        self.verbose(rf'Badge file found at {badge}')
                        badges.append((spdx, add_internal_asset(badge), uri))
            self.verbose_value(r'Context.license', self.license)

            # project repo access level
            self.private_repo = False
            if 'private_repo' in config:
                self.private_repo = bool(config['private_repo'])

            # project repository (github, gitlab, etc)
            self.repo = None
            for TYPE in repos.TYPES:
                if TYPE.KEY not in config:
                    continue
                self.verbose_value(rf'Context.{TYPE.KEY}', config[TYPE.KEY])
                try:
                    self.repo = TYPE(config[TYPE.KEY])
                except Error as err:
                    raise Error(rf'{TYPE.KEY}: {err}')
                self.verbose_value(rf'Context.repo', self.repo)
                if not self.private_repo and self.repo.release_badge_uri:
                    badges.append((r'Releases', self.repo.release_badge_uri, self.repo.releases_uri))

            # twitter
            self.twitter = None
            if r'twitter' in config and config[r'twitter']:
                self.twitter = config[r'twitter']
            self.verbose_value(r'Context.twitter', self.twitter)

            # sponsor
            self.sponsorship_uri = None
            if r'sponsor' in config and config[r'sponsor']:
                self.sponsorship_uri = config[r'sponsor']
            self.verbose_value(r'Context.sponsorship_uri', self.twitter)

            # project C++ version
            # defaults to 'current' cpp year version based on (current year - 2)
            # 1998, 2003, *range(2011, 2300, 3)
            default_cpp_year = max(int(self.now.year) - 2, 2011)
            default_cpp_year = default_cpp_year - ((default_cpp_year - 2011) % 3)
            self.cpp = default_cpp_year
            if r'cpp' in config:
                self.cpp = str(config['cpp']).lstrip('0 \t').rstrip()
                if not self.cpp:
                    self.cpp = default_cpp_year
                self.cpp = int(self.cpp)
                if self.cpp in (1998, 98):
                    self.cpp = 1998
                else:
                    if self.cpp > 2000:
                        self.cpp -= 2000
                    if self.cpp in [3, *range(11, 300, 3)]:
                        self.cpp += 2000
                    else:
                        raise Error(rf"cpp: '{config['cpp']}' is not a valid cpp standard version")
            self.verbose_value(r'Context.cpp', self.cpp)
            badge = rf'poxy-badge-c++{str(self.cpp)[2:]}.svg'
            badges.append(
                (rf'C++{str(self.cpp)[2:]}', rf'poxy/{badge}', r'https://en.cppreference.com/w/cpp/compiler_support')
            )
            add_internal_asset(badge)

            # project logo
            self.logo = None
            if r'logo' in config:
                if config['logo']:
                    file = config['logo'].strip()
                    if file:
                        file = Path(config['logo'])
                        if not file.is_absolute():
                            file = Path(self.input_dir, file)
                        self.logo = file.resolve()
            self.verbose_value(r'Context.logo', self.logo)

            # theme (HTML_EXTRA_STYLESHEETS, M_THEME_COLOR)
            self.theme = r'dark'
            if theme is not None:
                self.theme = theme
            elif r'theme' in config:
                self.theme = str(config[r'theme'])
            if self.theme != r'custom':
                self.stylesheets.append(add_internal_asset(paths.GENERATED / r'poxy.css'))
            self.verbose_value(r'Context.theme', self.theme)

            # stylesheets (HTML_EXTRA_STYLESHEETS)
            if r'stylesheets' in config:
                for f in coerce_collection(config[r'stylesheets']):
                    file = f.strip()
                    if file:
                        if is_uri(file):
                            self.stylesheets.append(file)
                        else:
                            file = Path(file)
                            self.stylesheets.append(file.name)
                            extra_files.append(file)
            self.verbose_value(r'Context.stylesheets', self.stylesheets)

            # jquery
            if r'jquery' in config and config[r'jquery']:
                jquery = enumerate_files(paths.JS, any=r'jquery*.js')[0]
                if jquery is not None:
                    self.scripts.append(add_internal_asset(jquery))

            # scripts
            self.scripts.append(add_internal_asset(paths.JS / r'poxy.js'))
            if r'scripts' in config:
                for f in coerce_collection(config[r'scripts']):
                    file = f.strip()
                    if file:
                        if is_uri(file):
                            self.scripts.append(file)
                        else:
                            file = Path(file)
                            self.scripts.append(file.name)
                            extra_files.append(file)
            self.verbose_value(r'Context.scripts', self.scripts)

            # enumerate blog files (need to add them to the doxygen sources)
            self.blog_files = []
            if self.blog_dir.exists() and self.blog_dir.is_dir():
                self.blog_files = enumerate_files(self.blog_dir, any=(r'*.md', r'*.markdown'), recursive=True)
                sep = re.compile(r'[-֊‐‑‒–—―−_ ,;.]+')
                expr = re.compile(
                    rf'^(?:blog{sep.pattern})?((?:[0-9]{{4}}){sep.pattern}(?:[0-9]{{2}}){sep.pattern}(?:[0-9]{{2}})){sep.pattern}[a-zA-Z0-9_ -]+$'
                )
                for i in range(len(self.blog_files)):
                    f = self.blog_files[i]
                    m = expr.fullmatch(f.stem)
                    if not m:
                        raise Error(
                            rf"blog post filename '{f.name}' was not formatted correctly; "
                            + r"it should be of the form 'YYYY-MM-DD_this_is_a_post.md'."
                        )
                    try:
                        d = datetime.datetime.strptime(sep.sub('-', m[1]), r'%Y-%m-%d').date()
                        self.blog_files[i] = (f, d)
                    except Exception as exc:
                        raise Error(rf"failed to parse date from blog post filename '{f.name}': {str(exc)}")
            self.verbose_value(r'Context.blog_files', self.blog_files)

            # changelog
            self.changelog = ''
            if r'changelog' in config:
                if isinstance(config['changelog'], bool):
                    if config['changelog']:
                        candidate_names = (r'CHANGELOG', r'CHANGES', r'HISTORY')
                        candidate_extensions = (r'.md', r'.txt', r'')
                        as_lowercase = (False, True)
                        candidate_dir = self.input_dir
                        while True:
                            for name, ext, lower in itertools.product(
                                candidate_names, candidate_extensions, as_lowercase
                            ):
                                candidate_file = Path(candidate_dir, rf'{name.lower() if lower else name}{ext}')
                                if (
                                    candidate_file.exists()
                                    and candidate_file.is_file()
                                    and candidate_file.stat().st_size <= 1024 * 1024 * 2
                                ):
                                    self.changelog = candidate_file
                                    break
                            if (
                                self.changelog
                                or candidate_dir.parent == candidate_dir
                                or (candidate_dir / '.git').exists()
                            ):
                                break
                            candidate_dir = candidate_dir.parent
                        if not self.changelog:
                            self.warning(
                                rf'changelog: Option was set to true but no file with a known changelog file name could be found! Consider using an explicit path.'
                            )

                else:
                    self.changelog = coerce_path(config['changelog'])
                    if not self.changelog.is_absolute():
                        self.changelog = Path(self.input_dir, self.changelog)
                    if not self.changelog.exists() or not self.changelog.is_file():
                        raise Error(rf'changelog: {config["changelog"]} did not exist or was not a file')
            if self.changelog:
                temp_changelog_path = Path(self.temp_pages_dir, r'poxy_changelog.md')
                copy_file(self.changelog, temp_changelog_path, logger=self.verbose_logger)
                self.changelog = temp_changelog_path
            self.verbose_value(r'Context.changelog', self.changelog)

            # main_page (USE_MDFILE_AS_MAINPAGE)
            self.main_page = ''
            if r'main_page' in config:
                if isinstance(config['main_page'], bool):
                    if config['main_page']:
                        candidate_names = (r'README', r'HOME', r'MAINPAGE', r'INDEX')
                        candidate_extensions = (r'.md', r'.txt', r'')
                        as_lowercase = (False, True)
                        candidate_dir = self.input_dir
                        while True:
                            for name, ext, lower in itertools.product(
                                candidate_names, candidate_extensions, as_lowercase
                            ):
                                candidate_file = Path(candidate_dir, rf'{name.lower() if lower else name}{ext}')
                                if (
                                    candidate_file.exists()
                                    and candidate_file.is_file()
                                    and candidate_file.stat().st_size <= 1024 * 1024 * 2
                                ):
                                    self.main_page = candidate_file
                                    break
                            if (
                                self.main_page
                                or candidate_dir.parent == candidate_dir
                                or (candidate_dir / '.git').exists()
                            ):
                                break
                            candidate_dir = candidate_dir.parent
                        if not self.main_page:
                            self.warning(
                                rf'main_page: Option was set to true but no file with a known main_page file name could be found! Consider using an explicit path.'
                            )

                else:
                    self.main_page = coerce_path(config['main_page'])
                    if not self.main_page.is_absolute():
                        self.main_page = Path(self.input_dir, self.main_page)
                    if not self.main_page.exists() or not self.main_page.is_file():
                        raise Error(rf'main_page: {config["main_page"]} did not exist or was not a file')
            self.verbose_value(r'Context.main_page', self.main_page)

            # sources (INPUT, FILE_PATTERNS, STRIP_FROM_PATH, STRIP_FROM_INC_PATH, EXTRACT_ALL)
            self.sources = Sources(
                config,
                r'sources',
                self.input_dir,
                additional_inputs=(
                    self.temp_pages_dir,  #
                    self.changelog if self.changelog else None,
                    *[f for f, d in self.blog_files],
                ),
                additional_strip_paths=(self.temp_pages_dir,),
            )
            self.verbose_object(r'Context.sources', self.sources)

            # images (IMAGE_PATH)
            self.images = Inputs(
                config,
                r'images',
                self.input_dir,
                additional_recursive_inputs=[self.blog_dir if self.blog_files else None],
            )
            self.verbose_object(r'Context.images', self.images)

            # examples (EXAMPLES_PATH, EXAMPLE_PATTERNS)
            self.examples = FilteredInputs(
                config,
                r'examples',
                self.input_dir,
                additional_recursive_inputs=[self.blog_dir if self.blog_files else None],
            )
            self.verbose_object(r'Context.examples', self.examples)

            # tagfiles (TAGFILES)
            self.tagfiles = {self.cppref_tagfile: (self.cppref_tagfile, r'http://en.cppreference.com/w/')}
            self.unresolved_tagfiles = False
            for k, v in extract_kvps(config, 'tagfiles').items():
                source = str(k)
                dest = str(v)
                if source and dest:
                    if is_uri(source):
                        file = Path(
                            paths.TEMP, rf'tagfile_{sha1(source)}_{self.now.year}_{self.now.isocalendar()[1]}.xml'
                        )
                        self.tagfiles[source] = (file, dest)
                        self.unresolved_tagfiles = True
                    else:
                        source = Path(source)
                        if not source.is_absolute():
                            source = Path(self.input_dir, source)
                        source = source.resolve()
                        self.tagfiles[str(source)] = (source, dest)
            for k, v in self.tagfiles.items():
                if isinstance(v, (Path, str)):
                    assert_existing_file(k)
            self.verbose_value(r'Context.tagfiles', self.tagfiles)

            # navbar
            if 1:
                # initialize
                self.navbar = []
                if r'navbar' in config:
                    for v in coerce_collection(config['navbar']):
                        val = v.strip()
                        if val:
                            self.navbar.append(val)
                else:
                    self.navbar = list(copy.deepcopy(Defaults.navbar))

                # expand 'default' and 'all'
                new_navbar = []
                for link in self.navbar:
                    if link == r'all':
                        new_navbar += [*Defaults.navbar_all]
                    elif link == r'default':
                        new_navbar += [*Defaults.navbar]
                    else:
                        new_navbar.append(link)
                self.navbar = new_navbar

                # normalize aliases
                for i in range(len(self.navbar)):
                    if self.navbar[i] == r'annotated':  # 'annotated' is doxygen-speak for 'classes'
                        self.navbar[i] = r'classes'
                    elif self.navbar[i] == r'modules':  # 'modules' is doxygen-speak for 'groups'
                        self.navbar[i] = r'groups'
                    elif self.navbar[i] == r'repository':
                        self.navbar[i] = r'repo'
                    elif self.navbar[i] in (r'sponsorship', r'funding', r'fund'):
                        self.navbar[i] = r'sponsor'

                # version switcher
                if self.versions_in_navbar and r'version' not in self.navbar:
                    self.navbar.append(r'version')
                while not self.versions_in_navbar and r'version' in self.navbar:
                    self.navbar.remove(r'version')

                # twitter
                if self.twitter and r'twitter' not in self.navbar:
                    self.navbar.append(r'twitter')
                while not self.twitter and r'twitter' in self.navbar:
                    self.navbar.remove(r'twitter')

                # repo logic
                if not self.repo:
                    for KEY in (r'repo', *repos.KEYS):
                        while KEY in self.navbar:
                            self.navbar.remove(KEY)
                else:
                    # remove repo buttons matching an uninstantiated repo type
                    for TYPE in repos.TYPES:
                        if not isinstance(self.repo, TYPE) and TYPE.KEY in self.navbar:
                            self.navbar.remove(TYPE.KEY)
                    # sub all remaining repo key aliases for simply 'repo'
                    for i in range(len(self.navbar)):
                        if self.navbar[i] in repos.KEYS:
                            self.navbar[i] = r'repo'
                    # add a repo button to the end if none was present
                    if r'repo' not in self.navbar:
                        self.navbar.append(r'repo')

                # sponsor
                if self.sponsorship_uri and r'sponsor' not in self.navbar:
                    self.navbar.append(r'sponsor')
                while not self.sponsorship_uri and r'sponsor' in self.navbar:
                    self.navbar.remove(r'sponsor')

                # theme logic
                if self.theme != r'custom' and r'theme' not in self.navbar:
                    self.navbar.append(r'theme')
                while self.theme == r'custom' and r'theme' in self.navbar:
                    self.navbar.remove(r'theme')

                # remove duplicates (working right-to-left)
                self.navbar.reverse()
                self.navbar = remove_duplicates(self.navbar)
                self.navbar.reverse()
                self.navbar = tuple(self.navbar)

                self.verbose_value(r'Context.navbar', self.navbar)

            # <meta> tags
            self.meta_tags = {}
            for k, v in extract_kvps(config, 'meta_tags', allow_blank_values=True).items():
                self.meta_tags[k] = v
            self.verbose_value(r'Context.meta_tags', self.meta_tags)

            # robots (<meta>)
            self.robots = True
            if 'robots' in config:
                self.robots = bool(config['robots'])
            self.verbose_value(r'Context.robots', self.robots)

            # inline namespaces for old versions of doxygen
            self.inline_namespaces = copy.deepcopy(Defaults.inline_namespaces)
            if 'inline_namespaces' in config:
                for ns in coerce_collection(config['inline_namespaces']):
                    namespace = ns.strip()
                    if namespace:
                        self.inline_namespaces.add(namespace)
            self.verbose_value(r'Context.inline_namespaces', self.inline_namespaces)

            self.excluded_symbols = set()
            if 'excluded_symbols' in config:
                for symbol in coerce_collection(config['excluded_symbols']):
                    symbol = symbol.strip()
                    if symbol:
                        self.excluded_symbols.add(symbol)

            # implementation headers
            self.implementation_headers = []
            if 'implementation_headers' in config:
                for k, v in config['implementation_headers'].items():
                    # header
                    header = k.strip().replace('\\', '/')
                    if not header:
                        continue
                    if header.find('*') != -1:
                        raise Error(rf"implementation_headers: target header path '{header}' may not have wildcards")
                    # impls
                    impls = coerce_collection(v)
                    impls = [i.strip().replace('\\', '/') for i in impls]
                    impls = [i for i in impls if i]
                    impls = sorted(remove_duplicates(impls))
                    if impls:
                        self.implementation_headers.append([header, impls])
            self.implementation_headers = tuple(self.implementation_headers)
            self.verbose_value(r'Context.implementation_headers', self.implementation_headers)

            # show_includes (SHOW_INCLUDES)
            self.show_includes = True
            if 'show_includes' in config:
                self.show_includes = bool(config['show_includes'])
            self.verbose_value(r'Context.show_includes', self.show_includes)

            # internal_docs (INTERNAL_DOCS)
            self.internal_docs = False
            if 'internal_docs' in config:
                self.internal_docs = bool(config['internal_docs'])
            self.verbose_value(r'Context.internal_docs', self.internal_docs)

            # generate_tagfile (GENERATE_TAGFILE)
            self.generate_tagfile = True
            self.tagfile_path = Path(
                self.temp_dir, rf'{self.name.replace(" ","_")}.tagfile.xml' if self.name else r'tagfile.xml'
            )
            if r'generate_tagfile' in config:
                self.generate_tagfile = bool(config[r'generate_tagfile'])
            self.verbose_value(r'Context.generate_tagfile', self.generate_tagfile)

            # favicon (M_FAVICON)
            self.favicon = None
            if 'favicon' in config:
                if config['favicon']:
                    file = Path(config['favicon'])
                    if not file.is_absolute():
                        file = Path(self.input_dir, file)
                    self.favicon = file.resolve()
                    extra_files.append(self.favicon)
            else:
                favicon = Path(self.input_dir, 'favicon.ico')
                if favicon.exists() and favicon.is_file():
                    self.favicon = favicon
                    extra_files.append(favicon)
            self.verbose_value(r'Context.favicon', self.favicon)

            # macros (PREDEFINED)
            self.macros = copy.deepcopy(Defaults.macros)
            for s in (r'defines', r'macros'):
                for k, v in extract_kvps(
                    config,
                    s,
                    value_getter=lambda v: (r'true' if v else r'false') if isinstance(v, bool) else str(v),
                    allow_blank_values=True,
                ).items():
                    self.macros[k] = v
            non_cpp_def_macros = copy.deepcopy(self.macros)
            cpp_defs = dict()
            for ver in [1998, 2003, *range(2011, 2300, 3)]:
                if ver > self.cpp:
                    break
                if ver not in Defaults.cpp_builtin_macros:
                    continue
                for k, v in Defaults.cpp_builtin_macros[ver].items():
                    cpp_defs[k] = v
            cpp_defs = [(k, v) for k, v in cpp_defs.items()]
            cpp_defs.sort(key=lambda kvp: kvp[0])
            for k, v in cpp_defs:
                self.macros[k] = v
            self.verbose_value(r'Context.macros', self.macros)

            # autolinks
            self.autolinks = [(k, v) for k, v in Defaults.autolinks.items()]
            if 'autolinks' in config:
                for pattern, u in config['autolinks'].items():
                    uri = u.strip()
                    if pattern.strip() and uri:
                        self.autolinks.append((pattern, uri))
            self.autolinks.sort(
                key=lambda v: (
                    self.__namespace_qualified.fullmatch(v[0]) is None,
                    v[0].find(r'std::') == -1,
                    -len(v[0]),
                    v[0],
                )
            )
            self.autolinks = tuple(self.autolinks)
            self.verbose_value(r'Context.autolinks', self.autolinks)

            # aliases (ALIASES)
            self.aliases = copy.deepcopy(Defaults.aliases)
            if 'aliases' in config:
                for k, v in config['aliases'].items():
                    alias = k.strip()
                    if not alias:
                        continue
                    if alias in self.aliases:
                        raise Error(rf'aliases.{k}: cannot override a built-in alias')
                        self.aliases[alias] = v
            self.verbose_value(r'Context.aliases', self.aliases)

            # badges for index.html banner
            user_badges = []
            if 'badges' in config:
                for k, v in config['badges'].items():
                    text = k.strip()
                    v = coerce_collection(v)
                    image_uri = v[0].strip()
                    anchor_uri = v[1].strip() if len(v) > 1 else r''
                    if text and image_uri:
                        user_badges.append((text, image_uri, anchor_uri))
            user_badges.sort(key=lambda b: b[0])
            self.badges = tuple(badges + user_badges)
            self.verbose_value(r'Context.badges', self.badges)

            # user-specified extra_files (HTML_EXTRA_FILES)
            if r'extra_files' in config:
                for f in coerce_collection(config['extra_files']):
                    file = f.strip()
                    if file:
                        extra_files.append(Path(file))

            # add all the 'icon' svgs as internal assets so they can be used by users if they wish
            for f in enumerate_files(paths.IMG, all='poxy-icon-*.svg', recursive=False):
                add_internal_asset(f)

            # finalize extra_files
            extra_files = remove_duplicates(extra_files)
            self.extra_files = {}
            for i in range(len(extra_files)):
                file = extra_files[i]
                if not isinstance(file, tuple):
                    path = coerce_path(file)
                    file = (path, path.name)
                else:
                    assert len(file) == 2
                    file = (coerce_path(file[0]), file[1])
                if not file[0].is_absolute():
                    file = (Path(self.input_dir, file[0]).resolve(), file[1])
                if not file[0].exists() or not file[0].is_file():
                    raise Error(rf'extra_files: {file[0]} did not exist or was not a file')
                if file[1] in self.extra_files:
                    raise Error(rf'extra_files: Multiple files with the name {file[1]}')
                self.extra_files[file[1]] = file[0]
            self.verbose_value(r'Context.extra_files', self.extra_files)

            # code_blocks
            self.code_blocks = CodeBlocks(config, non_cpp_def_macros)  # printed in run.py post-xml

            # html_header (HTML_HEADER in m.css)
            self.html_header = ''
            if r'html_header' in config:
                self.html_header = str(config[r'html_header']).strip()
            self.verbose_value(r'Context.html_header', self.html_header)

        # init emoji db
        self.emoji = emoji.Database()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.cleanup:
            delete_directory(self.temp_dir, logger=self.verbose_logger)

    def __bool__(self):
        return True


__all__ = ['Context']



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions for working with CSS files.
"""

from typing import Tuple

from . import paths
from .utils import *

RX_COMMENT = re.compile(r'''/[*].+?[*]/''', flags=re.DOTALL)
RX_IMPORT = re.compile(r'''@import\s+url\(\s*['"]?\s*(.+?)\s*['"]?\s*\)\s*;''', flags=re.I)
RX_MCSS_FILENAME = re.compile(r'(?:m|pygments)-[a-z0-9_-]+[.]css$', flags=re.I)
RX_GOOGLE_FONT = re.compile(
    r'''url\(\s*['"]?(https://fonts[.]gstatic[.]com/[a-z0-9_/%+?:-]+?[.]woff2)['"]?\s*\)''', flags=re.I
)


def strip_comments(text) -> str:
    global RX_COMMENT
    return RX_COMMENT.sub('', text)


def strip_quotes(text):
    if text.startswith(r'"') or text.startswith(r"'"):
        text = text[1:]
    if text.endswith(r'"') or text.endswith(r"'"):
        text = text[:-2]
    return text


def has_mcss_filename(path) -> bool:
    path = str(path).lower()
    if path.endswith(r'm-special.css'):
        return False

    global RX_MCSS_FILENAME
    return bool(RX_MCSS_FILENAME.match(path))


def resolve_imports(text, cwd=None, use_cached_fonts=True) -> Tuple[str, bool]:
    if cwd is None:
        cwd = Path.cwd()
    cwd = coerce_path(cwd).resolve()
    assert_existing_directory(cwd)

    had_mcss_files = False

    def match_handler(m):
        global RX_MCSS_THEME
        nonlocal cwd
        nonlocal had_mcss_files
        nonlocal use_cached_fonts

        import_path = strip_quotes(m[1].strip())
        had_mcss_files = had_mcss_files or has_mcss_filename(import_path)
        path = None
        path_ok = lambda: path is not None and path.exists() and path.is_file()

        # download + cache uris locally
        if is_uri(import_path):
            paths.GENERATED.mkdir(exist_ok=True)
            path = Path(paths.GENERATED, rf'{sha1(import_path.lower())}.css')
            if not path_ok() or (not use_cached_fonts and str(import_path).find(r'font') != -1):
                print(rf"Downloading {import_path}")
                data = download_text(import_path)
                print(rf"Writing {path}")
                with open(path, 'w', encoding='utf-8', newline='\n') as f:
                    f.write(data)

        # otherwise just check cwd
        if not path_ok():
            path = Path(cwd, import_path)

        # if we still haven't found a match just leave the @import statement as it was
        if not path_ok():
            return m[0]

        text = strip_comments(read_all_text_from_file(path, logger=True)).strip()
        header = rf'/*==== {import_path} {"="*(110-len(import_path))}*/'
        text = f'\n\n{header}\n{text}\n\n'

        res = resolve_imports(text, cwd=path.parent, use_cached_fonts=use_cached_fonts)
        had_mcss_files = had_mcss_files or res[1]
        return res[0]

    global RX_IMPORT
    return (RX_IMPORT.sub(match_handler, text), had_mcss_files)


def resolve_google_fonts(text, use_cached_fonts=True) -> str:
    def match_handler(m):
        global RX_GOOGLE_FONT
        uri = strip_quotes(m[1].strip())
        file_name = uri[uri.rfind('/') + 1 :]
        paths.FONTS.mkdir(exist_ok=True)
        path = Path(paths.FONTS, rf'{file_name}')
        if not path.exists() or not use_cached_fonts:
            print(rf"Downloading {uri}")
            data = download_binary(uri)
            print(rf"Writing {path}")
            with open(path, 'wb') as f:
                f.write(data)

        return rf"url('fonts/{file_name}')"

    global RX_GOOGLE_FONT
    return RX_GOOGLE_FONT.sub(match_handler, text)


def minify(text) -> str:
    text = text.replace('\t', ' ')
    text = re.sub(r'\n[ \t]*[{]', ' {\n', text)
    text = re.sub(r'[ \t]+\n', '\n', text)
    text = re.sub(r'\n\n+', '\n', text)

    pos = 0
    open_brackets = []
    while pos < len(text):
        if text[pos : pos + 2] == r'/*':
            pos += 2
            while pos < len(text) - 1:
                if text[pos : pos + 2] == r'*/':
                    pos += 2
                    break
                pos += 1
        else:
            if text[pos] == r'{':
                open_brackets.append(pos)
            elif text[pos] == r'}':
                if open_brackets:
                    block_content = text[open_brackets[-1] + 1 : pos].strip().replace('\n', ' ')
                    block_content = re.sub(r'[ \t][ \t]+', ' ', block_content)
                    if r'{' not in block_content and r'}' not in block_content:
                        semis = 0
                        for c in block_content:
                            if c == ';':
                                semis += 1
                            if semis >= 4:
                                break
                        if semis <= 4 and len(block_content) <= 100:
                            new_block = rf'{{ {block_content} }}'
                            text = text[: open_brackets[-1]] + new_block + text[pos + 1 :]
                            pos = open_brackets[-1] - 1 + len(new_block) - 1
                    open_brackets.pop()
            pos += 1

    text = re.sub(r'[{]\s+?[}]', r'{}', text)
    text = re.sub(r'[ \t][ \t]+', ' ', text)
    text = re.sub(r'\n[ \t]+([^ \t])', r'\n\1', text)
    return text


def regenerate_builtin_styles(use_cached_fonts=True):
    paths.GENERATED.mkdir(exist_ok=True)
    if not use_cached_fonts:
        delete_directory(paths.FONTS, logger=True)
    THEMES = (Path(paths.CSS, r'poxy.css'),)
    for theme_source_file in THEMES:
        text = strip_comments(read_all_text_from_file(theme_source_file, logger=True))
        text, had_mcss_files = resolve_imports(text, cwd=theme_source_file.parent, use_cached_fonts=use_cached_fonts)  #
        text = resolve_google_fonts(text, use_cached_fonts=use_cached_fonts)
        text = re.sub(r':(before|after)', r'::\1', text)
        text = re.sub(r':::+(before|after)', r'::\1', text)
        text = text.replace('\r\n', '\n')
        text = text.replace('\r', '\n')
        text = minify(text)
        if had_mcss_files:
            mcss_license = read_all_text_from_file(Path(paths.MCSS, 'COPYING'), logger=True).strip()
            text = rf'''/*
This file was automatically generated from multiple sources,
some of which included stylesheets from mosra/m.css.
The license for that project is as follows:

{mcss_license}
*/
{text}
'''
        theme_dest_file = Path(paths.GENERATED, theme_source_file.name)
        print(rf'Writing {theme_dest_file}')
        with open(theme_dest_file, r'w', encoding=r'utf-8', newline='\n') as f:
            f.write(text)


__all__ = [
    'strip_comments',
    'strip_quotes',
    'has_mcss_filename',
    'resolve_imports',
    'resolve_google_fonts',
    'minify',
    'regenerate_builtin_styles',
]



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions and types for working with C++ project graphs.
"""

import enum as _enum

from .utils import *

# =======================================================================================================================
# Node types
# =======================================================================================================================


class Namespace(object):
    '''Namespaces.'''

    pass


class Class(object):
    '''Classes.'''

    pass


class Struct(object):
    '''Structs.'''

    pass


class Union(object):
    '''Unions.'''

    pass


class Concept(object):
    '''C++20 `concept`.'''

    pass


class Function(object):
    '''Functions.'''

    pass


class Type(object):
    '''The type of a variable/enum/function return.'''

    pass


class Variable(object):
    '''Variables.'''

    pass


class Enum(object):
    '''Enums.'''

    pass


class EnumValue(object):
    '''Enum values.'''

    pass


class Typedef(object):
    '''Typedefs/aliases.'''

    pass


class Define(object):
    '''Preprocessor `#defines`'''

    pass


class Group(object):
    '''Groups (at the global level).'''

    pass


class MemberGroup(object):
    '''Member groups (at the class/union/struct level).'''

    pass


class Directory(object):
    '''A directory in the filesystem.'''

    pass


class File(object):
    '''A file.'''

    pass


class Page(object):
    '''A documentation page (e.g. Doxygen's `@page`).'''

    pass


class BriefDescription(object):
    '''A brief description of an element.'''

    pass


class DetailedDescription(object):
    '''A more detailed description of an element.'''

    pass


class Initializer(object):
    '''An initializer block (usually a code snippet).'''

    pass


class Paragraph(object):
    '''A paragraph.'''

    pass


class Text(object):
    '''Plain text.'''

    pass


class Reference(object):
    '''A reference to another node.'''

    pass


class ExternalResource(object):
    '''A reference to some resource outside the project (e.g. something in a tagfile).'''

    pass


class ExpositionMarkup(object):
    '''A 'leftover' node for representing miscellaneous markup in expository contexts.'''

    pass


class Friend(object):
    '''A friend relationship.'''

    pass


NODE_TYPES = {
    Namespace,
    Class,
    Struct,
    Union,
    Concept,
    Function,
    Variable,
    Enum,
    EnumValue,
    Typedef,
    Define,
    Group,
    MemberGroup,
    Directory,
    File,
    BriefDescription,
    DetailedDescription,
    Page,
    Initializer,
    Paragraph,
    Text,
    Reference,
    ExternalResource,
    Type,
    ExpositionMarkup,
    Friend,
}
DESCRIPTION_NODE_TYPES = {BriefDescription, DetailedDescription}
EXPOSITION_NODE_TYPES = {
    *DESCRIPTION_NODE_TYPES,
    Page,
    Initializer,
    Paragraph,
    Text,
    Reference,
    ExternalResource,
    Type,
    ExpositionMarkup,
}
CPP_TYPES = {Namespace, Class, Struct, Union, Concept, Function, Variable, Enum, EnumValue, Typedef, Define}
Namespace.CAN_CONTAIN = {Function, Class, Struct, Union, Variable, Typedef, Namespace, Concept, Enum, *DESCRIPTION_NODE_TYPES}
Class.CAN_CONTAIN = {Class, Struct, Union, Function, Variable, Typedef, Enum, MemberGroup, Friend, *DESCRIPTION_NODE_TYPES}
Struct.CAN_CONTAIN = Class.CAN_CONTAIN
Union.CAN_CONTAIN = Class.CAN_CONTAIN
Concept.CAN_CONTAIN = {Initializer, *DESCRIPTION_NODE_TYPES}
Function.CAN_CONTAIN = {Type, *DESCRIPTION_NODE_TYPES}
Variable.CAN_CONTAIN = {Type, Initializer, *DESCRIPTION_NODE_TYPES}
Enum.CAN_CONTAIN = {Type, EnumValue, *DESCRIPTION_NODE_TYPES}
EnumValue.CAN_CONTAIN = {Initializer, *DESCRIPTION_NODE_TYPES}
Typedef.CAN_CONTAIN = {Type, *DESCRIPTION_NODE_TYPES}
Define.CAN_CONTAIN = {Initializer, *DESCRIPTION_NODE_TYPES}
Group.CAN_CONTAIN = {t for t in NODE_TYPES if t not in (Page,)}
MemberGroup.CAN_CONTAIN = {t for t in Class.CAN_CONTAIN if t not in (MemberGroup,)}
Directory.CAN_CONTAIN = {Directory, File, *DESCRIPTION_NODE_TYPES}
File.CAN_CONTAIN = {
    Namespace,
    Class,
    Struct,
    Union,
    Concept,
    Function,
    Variable,
    Enum,
    Typedef,
    Define,
    Page,
    *DESCRIPTION_NODE_TYPES,
}
Page.CAN_CONTAIN = {Paragraph, Text, Reference, *DESCRIPTION_NODE_TYPES, ExpositionMarkup}
BriefDescription.CAN_CONTAIN = {Paragraph, Text, Reference, ExpositionMarkup}
DetailedDescription.CAN_CONTAIN = {Paragraph, Text, Reference, ExpositionMarkup}
Initializer.CAN_CONTAIN = {Text, Reference, ExpositionMarkup}
Paragraph.CAN_CONTAIN = {Text, Reference, ExpositionMarkup}
Text.CAN_CONTAIN = set()
ExpositionMarkup.CAN_CONTAIN = {Paragraph, Text, Reference, ExpositionMarkup}
Type.CAN_CONTAIN = {Text, Reference}
Reference.CAN_CONTAIN = {*CPP_TYPES, Page, Group, MemberGroup, Directory, File, ExternalResource}
ExternalResource.CAN_CONTAIN = set()
Friend.CAN_CONTAIN = {Function, Class, Struct, Union, *DESCRIPTION_NODE_TYPES}


@_enum.unique
class AccessLevel(_enum.Enum):
    PRIVATE = 0
    PROTECTED = 1
    PUBLIC = 2


# =======================================================================================================================
# Nodes
# =======================================================================================================================


def _make_node_iterator(nodes, *types):
    assert types is not None

    def permissive_generator():
        nonlocal nodes
        for node in nodes:
            yield node

    if not types:
        return permissive_generator()

    for t in types:
        assert t is None or isinstance(t, bool) or t in NODE_TYPES

    def selective_generator():
        nonlocal nodes
        nonlocal types
        yield_with_no_type = False in types or None in types
        yield_with_any_type = True in types
        for node in nodes:
            if (node.type is None and yield_with_no_type) or (
                node.type is not None and (yield_with_any_type or node.type in types)
            ):
                yield node

    return selective_generator()


class _NullNodeIterator(object):
    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration


class GraphError(Error):
    """Raised when a C++ graph error occurs."""

    pass


class GraphNodePropertyChanged(GraphError):
    """Raised when an attempt is made to change an already-set property in a graph node."""

    pass


class Node(object):
    """A single node in a C++ project graph."""

    class _Props(object):
        pass

    def __make_hierarchy_containers(self):
        if hasattr(self, r'_Node__children'):
            return
        self.__parents = []
        self.__parents_by_id = dict()
        self.__children = []
        self.__children_by_id = dict()

    def __init__(self, id: str):
        assert id is not None
        self.__id = id

    # ==============
    # getters
    # ==============

    def __property_get(self, name: str, out_type=None, default=None):
        assert name is not None
        value = None
        if hasattr(self, r'_Node__props'):
            value = getattr(self.__props, str(name), None)
        if value is None:
            value = default
        if value is not None and out_type is not None and not isinstance(value, out_type):
            value = out_type(value)
        return value

    @property
    def id(self) -> str:
        return self.__id

    @property
    def type(self):
        return self.__property_get(r'type')

    @property
    def type_name(self) -> str:
        t = self.type
        if t is None:
            return ''
        return t.__name__

    @property
    def qualified_name(self) -> str:
        return self.__property_get(r'qualified_name', str, r'')

    @property
    def local_name(self) -> str:
        return self.__property_get(r'local_name', str, r'')

    @property
    def definition(self) -> str:
        return self.__property_get(r'definition', str, r'')

    @property
    def static(self) -> bool:
        return self.__property_get(r'static', bool, False)

    @property
    def const(self) -> bool:
        return self.__property_get(r'const', bool, False)

    @property
    def constexpr(self) -> bool:
        return self.__property_get(r'constexpr', bool, False)

    @property
    def constinit(self) -> bool:
        return self.__property_get(r'constinit', bool, False)

    @property
    def consteval(self) -> bool:
        return self.__property_get(r'consteval', bool, False)

    @property
    def inline(self) -> bool:
        return self.__property_get(r'inline', bool, False)

    @property
    def final(self) -> bool:
        return self.__property_get(r'final', bool, False)

    @property
    def explicit(self) -> bool:
        return self.__property_get(r'explicit', bool, False)

    @property
    def noexcept(self) -> bool:
        return self.__property_get(r'noexcept', bool, False)

    @property
    def virtual(self) -> bool:
        return self.__property_get(r'virtual', bool, False)

    @property
    def mutable(self) -> bool:
        return self.__property_get(r'mutable', bool, False)

    @property
    def strong(self) -> bool:
        return self.__property_get(r'strong', bool, False)

    @property
    def access_level(self) -> AccessLevel:
        return self.__property_get(
            r'access_level', AccessLevel, AccessLevel.PRIVATE if self.type is Class else AccessLevel.PUBLIC
        )

    @property
    def text(self) -> str:
        return self.__property_get(r'text', str, r'')

    @property
    def is_paragraph(self) -> bool:
        return self.__property_get(r'is_paragraph', bool, False)

    @property
    def file(self) -> str:
        return self.__property_get(r'file', str, r'')

    @property
    def line(self) -> int:
        return self.__property_get(r'line', int, 0)

    @property
    def column(self) -> int:
        return self.__property_get(r'column', int, 0)

    @property
    def kind(self) -> str:
        return self.__property_get(r'kind', str, r'')

    @property
    def tag(self) -> str:
        return self.__property_get(r'tag', str, r'')

    @property
    def extra_attributes(self) -> typing.Sequence[typing.Tuple[str, str]]:
        return self.__property_get(r'extra_attributes', None, tuple())

    def __bool__(self) -> bool:
        return self.type is not None and bool(self.id)

    # ==============
    # setters
    # ==============

    def __property_set(self, name: str, out_type, value, strip_strings=False):
        assert name is not None
        # known types that have a sensible __bool__ operator can convert to None if false
        if isinstance(value, (str, Path, list, tuple, dict)):
            value = value if value else None
        # converting from strings sometimes lets us do some light parsing, as a treat
        if isinstance(value, str):
            if out_type is bool:
                if value.lower() in (r'no', r'false', r'disabled'):
                    value = False
                elif value.lower() in (r'yes', r'true', r'enabled'):
                    value = True
                else:
                    raise GraphError(rf"Node '{self.id}' property '{name}' could not parse a boolean from '{value}'")
            elif out_type is AccessLevel:
                if value.lower() in (r'pub', r'public'):
                    value = AccessLevel.PUBLIC
                elif value.lower() in (r'prot', r'protected'):
                    value = AccessLevel.PROTECTED
                elif value.lower() in (r'priv', r'private'):
                    value = AccessLevel.PRIVATE
                else:
                    raise GraphError(rf"Node '{self.id}' property '{name}' could not parse access level from '{value}'")
                assert isinstance(value, AccessLevel)
        # None == keep whatever the current value is (no-op)
        # (None is never a valid value for a real graph attribute)
        if value is None:
            return
        if out_type is not None and not isinstance(value, out_type):
            value = out_type(value)
        if strip_strings and isinstance(value, str):
            value = value.strip()
        current = None
        has_props = hasattr(self, r'_Node__props')
        if has_props:
            current = getattr(self.__props, str(name), None)
        # it's OK if there's already a value as long as it's identical to the new one,
        # otherwise we throw so that we can detect when the source data is bad or the adapter is faulty
        # (since if a property _can_ be defined in multiple places it should be identical in all of them)
        if current is not None:
            if type(current) != type(value):
                raise GraphNodePropertyChanged(
                    rf"Node '{self.id}' property '{name}' first seen with type {type(current)}, now seen with type {type(value)}"
                )
            if current != value:
                raise GraphNodePropertyChanged(
                    rf"Node '{self.id}' property '{name}' first seen with value '{current}', now seen with value '{value}'"
                )
            return
        if not has_props:
            self.__props = Node._Props()
        setattr(self.__props, str(name), value)

    @type.setter
    def type(self, value):
        if value is None:
            return
        if value not in NODE_TYPES:
            raise GraphError(rf"Unknown C++ node type '{value}'")
        had_type = self.type is not None
        self.__property_set(r'type', None, value)
        if had_type != (self.type is not None):
            self.__deduce_local_name()
            if hasattr(self, r'_Node__children'):
                for child in self.__children:
                    Node._check_connection(self, child)

    def __deduce_local_name(self):
        if not self.qualified_name or self.local_name or self.type is None:
            return
        if self.type in (Namespace, Class, Struct, Union, Concept, Function, Variable, Enum, EnumValue, Typedef):
            if self.qualified_name.find(r'<') != -1:  # templates might have template args with '::' so ignore them
                return
            self.local_name = tail(self.qualified_name, r'::')
        elif self.type in (Directory, File):
            self.local_name = tail(self.qualified_name, r'/')
        elif self.type is Define:
            self.local_name = self.qualified_name

    @qualified_name.setter
    def qualified_name(self, value: str):
        if value is not None and self.type in (Directory, File):
            value = str(value).strip().replace('\\', r'/').rstrip(r'/')
        self.__property_set(r'qualified_name', str, value, strip_strings=True)
        self.__deduce_local_name()

    @local_name.setter
    def local_name(self, value: str):
        self.__property_set(r'local_name', str, value, strip_strings=True)

    @definition.setter
    def definition(self, value: str):
        self.__property_set(r'definition', str, value)

    @static.setter
    def static(self, value: bool):
        self.__property_set(r'static', bool, value)

    @const.setter
    def const(self, value: bool):
        self.__property_set(r'const', bool, value)

    @constexpr.setter
    def constexpr(self, value: bool):
        self.__property_set(r'constexpr', bool, value)

    @constinit.setter
    def constinit(self, value: bool):
        self.__property_set(r'constinit', bool, value)

    @consteval.setter
    def consteval(self, value: bool):
        self.__property_set(r'consteval', bool, value)

    @inline.setter
    def inline(self, value: bool):
        self.__property_set(r'inline', bool, value)

    @final.setter
    def final(self, value: bool):
        self.__property_set(r'final', bool, value)

    @explicit.setter
    def explicit(self, value: bool):
        self.__property_set(r'explicit', bool, value)

    @noexcept.setter
    def noexcept(self, value: bool):
        self.__property_set(r'noexcept', bool, value)

    @virtual.setter
    def virtual(self, value: bool):
        self.__property_set(r'virtual', bool, value)

    @mutable.setter
    def mutable(self, value: bool):
        self.__property_set(r'mutable', bool, value)

    @strong.setter
    def strong(self, value: bool):
        self.__property_set(r'strong', bool, value)

    @access_level.setter
    def access_level(self, value: AccessLevel):
        self.__property_set(r'access_level', AccessLevel, value)

    @text.setter
    def text(self, value: str):
        self.__property_set(r'text', str, value)

    @is_paragraph.setter
    def is_paragraph(self, value: bool):
        self.__property_set(r'is_paragraph', bool, value)

    @file.setter
    def file(self, value: str):
        if value is not None:
            value = str(value).strip().replace('\\', r'/').rstrip(r'/')
        self.__property_set(r'file', str, value)

    @line.setter
    def line(self, value: int):
        self.__property_set(r'line', int, value)

    @column.setter
    def column(self, value: int):
        self.__property_set(r'column', int, value)

    @kind.setter
    def kind(self, value: str):
        self.__property_set(r'kind', str, value, strip_strings=True)

    @tag.setter
    def tag(self, value: str):
        self.__property_set(r'tag', str, value, strip_strings=True)

    @extra_attributes.setter
    def extra_attributes(self, value: typing.Sequence[typing.Tuple[str, str]]):
        self.__property_set(r'extra_attributes', None, value)

    # ==============
    # children
    # ==============

    @property
    def is_child(self) -> bool:
        return hasattr(self, r'_Node__children') and bool(self.__parents)

    @property
    def is_parent(self) -> bool:
        return hasattr(self, r'_Node__children') and bool(self.__children)

    def __contains__(self, node_or_id) -> bool:
        assert node_or_id is not None
        assert isinstance(node_or_id, (str, Node)) or node_or_id in NODE_TYPES
        if not hasattr(self, r'_Node__children'):
            return False
        if isinstance(node_or_id, Node):
            node_or_id = node_or_id.id
        if isinstance(node_or_id, str):
            return node_or_id in self.__children_by_id
        else:
            for c in self.__children:
                if c.type is node_or_id:
                    return True
            return False

    def has_parent(self, *node_or_id_or_types) -> bool:
        assert node_or_id_or_types is not None
        if not hasattr(self, r'_Node__parents'):
            return False
        for target in node_or_id_or_types:
            if isinstance(target, Node):
                target = target.id
            if isinstance(target, str):
                return target in self.__parents_by_id
            else:
                assert target in NODE_TYPES
                for c in self.__parents:
                    if c.type is target:
                        return True
        return False

    def __getitem__(self, id_or_index: typing.Union[str, int]):
        assert id_or_index is not None
        assert isinstance(id_or_index, (str, int)) or id_or_index in NODE_TYPES
        if not hasattr(self, r'_Node__children'):
            return None
        if isinstance(id_or_index, str):
            try:
                return self.__children_by_id[id_or_index]
            except:
                return None
        elif isinstance(id_or_index, int):
            return self.__children[id_or_index]
        else:
            for c in self(id_or_index):
                return c
            raise KeyError(id_or_index.__name__)

    @classmethod
    def _check_connection(cls, source, dest):
        assert source is not None
        assert isinstance(source, Node)
        assert dest is not None
        assert isinstance(dest, Node)

        # self-connection is always illegal, regardless of type information
        if id(source) == id(dest):
            raise GraphError(rf"Node '{source.id}' may not connect to itself")

        # otherwise if we don't have type information the connection is 'OK'
        # (really this just means we defer the check until later)
        if source.type is None or dest.type is None:
            return

        # check basic connection rules
        if dest.type not in source.type.CAN_CONTAIN:
            raise GraphError(rf"{source.type_name} node '{source.id}' is not allowed to connect to {dest.type_name} nodes")

        # check situations where a node must only belong to one parent of a particular set of types
        def check_single_parent(dest_types, source_types):
            nonlocal source
            nonlocal dest
            source_types = coerce_collection(source_types)
            dest_types = coerce_collection(dest_types)
            if source.type not in source_types or dest.type not in dest_types:
                return
            sum = 0
            for parent in dest(*source_types, parents=True):
                sum += 1
            if dest not in source:
                sum += 1
            if sum > 1:
                raise GraphError(
                    rf"{dest.type_name} node '{dest.id}' is not allowed to be a member of more than one "
                    + rf"{{ {', '.join([t.__name__ for t in source_types])} }}"
                )

        check_single_parent(EnumValue, Enum)
        check_single_parent(Type, (Variable, Function, Enum, Typedef))
        check_single_parent(Friend, (Class, Struct, Union))
        check_single_parent(Reference, NODE_TYPES)

        # same again but in the other direction
        def check_single_child(source_types, dest_types):
            nonlocal source
            nonlocal dest
            source_types = coerce_collection(source_types)
            dest_types = coerce_collection(dest_types)
            if source.type not in source_types or dest.type not in dest_types:
                return
            sum = 0
            for child in source(*dest_types):
                sum += 1
            if dest not in source:
                sum += 1
            if sum > 1:
                raise GraphError(
                    rf"{source.type_name} node '{source.id}' is not allowed to be connected to more than one "
                    + rf"{{ {', '.join([t.__name__ for t in dest_types])} }}"
                )

        check_single_child(Friend, (Class, Struct, Union, Function))
        check_single_child(Reference, NODE_TYPES)

    def add(self, child):
        assert child is not None
        assert isinstance(child, Node)

        # connecting to the same node twice is fine (no-op)
        if child in self:
            existing_child = self.__children_by_id[child.id]
            # check that identity is unique
            if id(child) != id(existing_child):
                raise GraphError(rf"Two different nodes seen with the same ID ('{child.id}')")
            return

        Node._check_connection(self, child)

        self.__make_hierarchy_containers()
        self.__children.append(child)
        self.__children_by_id[child.id] = child

        child.__make_hierarchy_containers()
        child.__parents.append(self)
        child.__parents_by_id[self.id] = self

    def __iter__(self):
        if not hasattr(self, r'_Node__children'):
            return _NullNodeIterator()
        return _make_node_iterator(self.__children)

    def __call__(self, *types, parents=False):
        if not hasattr(self, r'_Node__children'):
            return _NullNodeIterator()
        return _make_node_iterator(self.__parents if parents else self.__children, *types)

    def remove(self, child):
        assert child is not None
        assert isinstance(child, Node)

        if not hasattr(self, r'_Node__children') or child not in self or child is self:
            return

        self.__children.remove(child)
        del self.__children_by_id[child.id]

        child.__parents.remove(self)
        del child.__parents_by_id[self.id]

    def clear(self):
        if not hasattr(self, r'_Node__children'):
            return

        for child in self.__children:
            child.__parents.remove(self)
            del child.__parents_by_id[self.id]

        self.__children.clear()
        self.__children_by_id.clear()

    # ==============
    # relationship queries
    # ==============

    @property
    def is_class_member(self) -> bool:
        return self.has_parent(Class, Struct, Union)

    @property
    def is_class_member_variable(self) -> bool:
        return self.type is Variable and self.is_class_member

    @property
    def is_class_member_variable(self) -> bool:
        return self.type is Function and self.is_class_member

    @property
    def is_free_function(self) -> bool:
        return self.type is Function and not self.is_class_member

    @property
    def is_static_function(self) -> bool:
        return self.type is Function and self.static

    @property
    def is_friend(self) -> bool:
        return self.has_parent(Friend)

    @property
    def is_friend_function(self) -> bool:
        return self.type is Function and self.is_friend

    @property
    def is_friend_class(self) -> bool:
        return self.type in (Class, Struct, Union) and self.is_friend

    @property
    def has_friends(self) -> bool:
        return Friend in self

    # ==============
    # misc
    # ==============

    def copy(self, id=None, transform=None):
        node = Node(self.id if id is None else id)
        if transform is not None:
            transform(self, node)
        if hasattr(self, r'_Node__props'):
            node.__props = Node._Props()
            for key, val in self.__props.__dict__.items():
                if not hasattr(node.__props, key):
                    setattr(node.__props, key, val)
        return node


# =======================================================================================================================
# Graph
# =======================================================================================================================


class Graph(object):
    """A C++ project graph."""

    def __init__(self):
        self.__nodes: typing.Dict[str, Node]
        self.__nodes = dict()
        self.__next_unique_id = 0

    def __get_unique_id(self) -> str:
        id = rf'__graph_unique_id_{self.__next_unique_id}'
        self.__next_unique_id += 1
        return id

    def get_or_create_node(self, id: str = None, type=None, parent=None) -> Node:
        if id is None:
            id = self.__get_unique_id()
        assert id
        node = None
        if id not in self.__nodes:
            node = Node(id)
            self.__nodes[id] = node
        else:
            node = self.__nodes[id]
        node.type = type
        if parent is not None:
            parent.add(node)
        return node

    def __iter__(self):
        return _make_node_iterator(self.__nodes.values())

    def __call__(self, *types):
        return _make_node_iterator(self.__nodes.values(), *types)

    def __contains__(self, node_or_id) -> bool:
        assert node_or_id is not None
        assert isinstance(node_or_id, (str, Node)) or node_or_id in NODE_TYPES
        if isinstance(node_or_id, Node):
            node_or_id = node_or_id.id
        if isinstance(node_or_id, str):
            return node_or_id in self.__nodes
        else:
            for _, n in self.__nodes:
                if n.type is node_or_id:
                    return True
            return False

    def __getitem__(self, id: str) -> Node:
        assert id is not None
        assert isinstance(id, str)
        try:
            return self.__nodes[id]
        except:
            return None

    def remove(self, *nodes: typing.Sequence[Node], filter=None):
        if filter is not None and not nodes:
            nodes = self.__nodes.values()
        prune = []
        for node in nodes:
            if node is None or node not in self:
                continue
            if filter is not None and not filter(node):
                continue
            for _, other in self.__nodes.items():
                if node is not other:
                    other.remove(node)
            node.clear()
            prune.append(node)
        for node in prune:
            del self.__nodes[node.id]

    def validate(self):
        for node in self:
            if node.type is None:
                raise GraphError(rf"Node '{node.id}' is untyped")
            if node.type not in EXPOSITION_NODE_TYPES:
                if not node.qualified_name:
                    raise GraphError(rf"{node.type_name} node '{node.id}' missing attribute 'qualified_name'")
                if not node.local_name:
                    raise GraphError(rf"{node.type_name} node '{node.id}' missing attribute 'local_name'")

            if node.file.find('\\') != -1:
                raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'file' contains back-slashes")
            if node.file.endswith(r'/'):
                raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'file' ends with a forward-slash")
            if node.line < 0:
                raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'line' is negative")
            if node.column < 0:
                raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'column' is negative")

            if node.type in (Directory, File):
                if node.qualified_name.find('\\') != -1:
                    raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'qualified_name' contains back-slashes")
                if node.qualified_name.endswith(r'/'):
                    raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'qualified_name' ends with a forward-slash")
            if node.type in CPP_TYPES:
                if node.qualified_name.startswith(r'::'):
                    raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'qualified_name' starts with ::")
                if node.qualified_name.endswith(r'::'):
                    raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'qualified_name' ends with ::")
                if node.type is not EnumValue and not node.file:
                    raise GraphError(rf"{node.type_name} node '{node.id}' missing attribute 'file'")

            if node.type in (EnumValue, Type, Friend):
                if not node.is_child:
                    raise GraphError(rf"{node.type_name} node '{node.id}' is an orphan")

            if node.type in (Function, Variable, Typedef):
                if Type not in node:
                    raise GraphError(rf"{node.type_name} node '{node.id}' is missing a Type")

    def copy(self, filter=None, id_transform=None, transform=None):
        g = Graph()
        id_remap = dict()
        # first pass to copy
        for src in self:
            if filter is not None and not filter(src):
                continue
            id = src.id
            if id_transform is not None:
                id = id_transform(src)
            if id is None:
                id = g.__get_unique_id()
            else:
                id = str(id)
            if id in g:
                raise GraphError(rf"A node with id '{id}' already exists in the destination graph")
            id_remap[src.id] = id
            g.__nodes[id] = src.copy(id=id, transform=transform)
        # second pass to link hierarchy
        for src in self:
            if src.id not in id_remap:
                continue
            for child in src:
                if child.id not in id_remap:
                    continue
                g[id_remap[src.id]].add(g[id_remap[child.id]])
        return g



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import re
import subprocess

_patch_src = re.compile(r"""<\?xml version="1\.0" encoding="UTF-8" standalone="no"\?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1\.1//EN"
 "http://www\.w3\.org/Graphics/SVG/1\.1/DTD/svg11\.dtd">
<svg width="(?P<width>\d+)pt" height="(?P<height>\d+)pt"
 viewBox="(?P<viewBox>[^"]+)" xmlns="http://www\.w3\.org/2000/svg" xmlns:xlink="http://www\.w3\.org/1999/xlink">
<g id="graph0" class="graph" """)

_patch_dst = r"""<svg{attribs} style="width: {width:.3f}rem; height: {height:.3f}rem;" viewBox="{viewBox}">
<g """
_patch_custom_size_dst = r"""<svg{attribs} style="{size}" viewBox="\g<viewBox>">
<g """

_comment_src = re.compile(r"""<!--[^-]+-->\n""")

_class_src = re.compile(r"""<g id="(edge|node|clust)\d+" class="(?P<type>edge|node|cluster)(?P<classes>[^"]*)">
<title>(?P<title>[^<]*)</title>
<(?P<element>ellipse|polygon|path|text)( fill="(?P<fill>[^"]+)" stroke="[^"]+")? """)

_class_dst = r"""<g class="{classes}">
<title>{title}</title>
<{element} """

_attributes_src = re.compile(r"""<(?P<element>ellipse|polygon|polyline) fill="[^"]+" stroke="[^"]+" """)

_attributes_dst = r"""<\g<element> """

# re.compile() is called after replacing {font} in configure(). The fill
# attribute is sometimes omitted.
_text_src_src = ' font-family="{font}" font-size="(?P<size>[^"]+)"( fill="[^"]+")?'

_text_dst = ' style="font-size: {size}px;"'

_font = ''
_font_size = 0.0

# The pt are actually px (16pt font is the same size as 16px), so just
# converting to rem here
def _pt2em(pt): return pt/_font_size

def dot2svg(source, size=None, attribs=''):
    try:
        ret = subprocess.run(['dot', '-Tsvg',
            '-Gfontname={}'.format(_font),
            '-Nfontname={}'.format(_font),
            '-Efontname={}'.format(_font),
            '-Gfontsize={}'.format(_font_size),
            '-Nfontsize={}'.format(_font_size),
            '-Efontsize={}'.format(_font_size),
            '-Gbgcolor=transparent',
            ], input=source.encode('utf-8'), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if ret.returncode: print(ret.stderr.decode('utf-8'))
        ret.check_returncode()
    except FileNotFoundError: # pragma: no cover
        raise RuntimeError("dot not found")

    # First remove comments
    svg = _comment_src.sub('', ret.stdout.decode('utf-8'))

    # Remove preamble and fixed size
    if size:
        svg = _patch_src.sub(_patch_custom_size_dst.format(attribs=attribs, size=size), svg)
    else:
        def patch_repl(match): return _patch_dst.format(
            attribs=attribs,
            width=_pt2em(float(match.group('width'))),
            height=_pt2em(float(match.group('height'))),
            viewBox=match.group('viewBox'))
        svg = _patch_src.sub(patch_repl, svg)

    # Remove unnecessary IDs and attributes, replace classes for elements
    def element_repl(match):
        classes = ['m-' + match.group('type')] + match.group('classes').replace('&#45;', '-').split()
        # distinguish between solid and filled nodes
        if ((match.group('type') == 'node' and match.group('fill') == 'none') or
            # a plaintext node is also flat
            match.group('element') == 'text'
        ):
            classes += ['m-flat']

        return _class_dst.format(
            classes=' '.join(classes),
            title=match.group('title'),
            element=match.group('element'))
    svg = _class_src.sub(element_repl, svg)

    # Remove unnecessary fill and stroke attributes
    svg = _attributes_src.sub(_attributes_dst, svg)

    # Remove unnecessary text attributes. Keep font size only if nondefault
    def text_repl(match):
        if float(match.group('size')) != _font_size:
            return _text_dst.format(size=float(match.group('size')))
        return ''
    svg = _text_src.sub(text_repl, svg)

    return svg

def configure(font, font_size):
    global _font, _font_size, _text_src
    _font = font
    _font_size = font_size
    _text_src = re.compile(_text_src_src.format(font=_font))
